"""convert timestamps to with timezones

Revision ID: 42ea01a6cf31
Revises: ce05198f524a
Create Date: 2021-06-26 23:50:50.271191

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '42ea01a6cf31'
down_revision = 'ce05198f524a'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('agents', 'created',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('agents', 'deleted',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('agents', 'last_seen',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('agents', 'updated',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('agents_groups', 'created',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('agents_groups', 'deleted',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('agents_groups', 'updated',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('branches', 'created',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('branches', 'deleted',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('branches', 'updated',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('flows', 'created',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('flows', 'deleted',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('flows', 'finished',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('flows', 'updated',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('jobs', 'assigned',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('jobs', 'completed',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('jobs', 'created',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('jobs', 'deleted',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('jobs', 'finished',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('jobs', 'processing_started',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('jobs', 'started',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('jobs', 'updated',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('projects', 'created',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('projects', 'deleted',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('projects', 'updated',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('repo_changes', 'created',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('repo_changes', 'deleted',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('repo_changes', 'updated',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('runs', 'created',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('runs', 'deleted',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('runs', 'email_sent',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('runs', 'finished',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('runs', 'finished_again',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('runs', 'hard_timeout_reached',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('runs', 'soft_timeout_reached',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('runs', 'started',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('runs', 'updated',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('secrets', 'created',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('secrets', 'deleted',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('secrets', 'updated',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('stages', 'created',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('stages', 'deleted',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('stages', 'updated',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('steps', 'created',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('steps', 'deleted',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('steps', 'updated',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('test_cases', 'created',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('test_cases', 'deleted',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('test_cases', 'updated',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('tools', 'created',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('tools', 'deleted',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('tools', 'updated',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('user_sessions', 'created',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('user_sessions', 'deleted',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('user_sessions', 'updated',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('users', 'created',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    op.alter_column('users', 'deleted',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=True)
    op.alter_column('users', 'updated',
               existing_type=postgresql.TIMESTAMP(),
               type_=sa.DateTime(timezone=True),
               existing_nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('users', 'updated',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('users', 'deleted',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('users', 'created',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('user_sessions', 'updated',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('user_sessions', 'deleted',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('user_sessions', 'created',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('tools', 'updated',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('tools', 'deleted',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('tools', 'created',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('test_cases', 'updated',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('test_cases', 'deleted',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('test_cases', 'created',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('steps', 'updated',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('steps', 'deleted',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('steps', 'created',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('stages', 'updated',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('stages', 'deleted',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('stages', 'created',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('secrets', 'updated',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('secrets', 'deleted',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('secrets', 'created',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('runs', 'updated',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('runs', 'started',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('runs', 'soft_timeout_reached',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('runs', 'hard_timeout_reached',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('runs', 'finished_again',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('runs', 'finished',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('runs', 'email_sent',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('runs', 'deleted',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('runs', 'created',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('repo_changes', 'updated',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('repo_changes', 'deleted',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('repo_changes', 'created',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('projects', 'updated',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('projects', 'deleted',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('projects', 'created',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('jobs', 'updated',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('jobs', 'started',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('jobs', 'processing_started',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('jobs', 'finished',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('jobs', 'deleted',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('jobs', 'created',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('jobs', 'completed',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('jobs', 'assigned',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('flows', 'updated',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('flows', 'finished',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('flows', 'deleted',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('flows', 'created',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('branches', 'updated',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('branches', 'deleted',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('branches', 'created',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('agents_groups', 'updated',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('agents_groups', 'deleted',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('agents_groups', 'created',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('agents', 'updated',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    op.alter_column('agents', 'last_seen',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('agents', 'deleted',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=True)
    op.alter_column('agents', 'created',
               existing_type=sa.DateTime(timezone=True),
               type_=postgresql.TIMESTAMP(),
               existing_nullable=False)
    # ### end Alembic commands ###



"""added git_clone_params to stage

Revision ID: 54efe873f88f
Revises: e0d8421619c2
Create Date: 2021-09-15 06:40:12.583695

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '54efe873f88f'
down_revision = 'e0d8421619c2'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('stages', sa.Column('git_clone_params', sa.UnicodeText(), nullable=True))


def downgrade():
    op.drop_column('stages', 'git_clone_params')



#!/usr/bin/env python3

# Copyright 2020-2021 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging

import connexion
from connexion.resolver import Resolver
from flask import request

from . import logs
from . import models
from . import backend
from . import consts
from . import srvcheck
from . import webhooks
from . import agentblob
from . import storage
from . import job_log
from . import badge
from . import minioops
from . import access
from . import authn
from .. import version

log = logging.getLogger('server')


class MyResolver(Resolver):
    def resolve_operation_id(self, operation):
        operation_id = operation.operation_id
        tags = operation._operation['tags']  # pylint: disable=protected-access
        name = 'kraken.server.{}.{}'.format(tags[0].lower(), operation_id)
        return name


def _set_log_ctx():
    if request.path.startswith(('/api', '/bk/api')):
        name = 'api'
    elif request.path.startswith(('/backend', '/bk/backend')):
        name = 'backend'
    elif request.path.startswith(('/install', '/bk/install')):
        name = 'install'
    elif request.path.startswith(('/artifacts', '/bk/artifacts')):
        name = 'artifacts'
    elif request.path.startswith(('/job_log', '/bk/job_log')):
        name = 'job-log'
    elif request.path.startswith(('/branch-badge', '/bk/branch-badge')):
        name = 'badge'
    elif request.path.startswith(('/webhooks', '/bk/webhooks')):
        name = 'webhooks'
    else:
        name = 'other'

    try:
        log.set_ctx(tool=name)
    except Exception:
        pass


def _clear_request_ctx(a):  # pylint: disable=unused-argument
    try:
        log.set_ctx(tool=None)
    except Exception:
        log.exception('IGNORED')

    try:
        models.db.session.remove()
    except Exception:
        log.exception('IGNORED')


def _unhandled_error_handler(err):
    log.info('ERROR %s', err)
    print('ERROR', err)
    log.info('ORIG ERROR %s', err.original_exception)
    print('ORIG ERROR', err.original_exception)
    return '{}', 500


def create_app():
    # addresses
    db_url = os.environ.get('KRAKEN_DB_URL', consts.DEFAULT_DB_URL)
    redis_addr = os.environ.get('KRAKEN_REDIS_ADDR', consts.DEFAULT_REDIS_ADDR)
    clickhouse_url = os.environ.get('KRAKEN_CLICKHOUSE_URL', consts.DEFAULT_CLICKHOUSE_URL)
    planner_url = os.environ.get('KRAKEN_PLANNER_URL', consts.DEFAULT_PLANNER_URL)
    server_addr = os.environ.get('KRAKEN_SERVER_ADDR', consts.DEFAULT_SERVER_ADDR)

    srvcheck.check_postgresql(db_url)
    srvcheck.wait_for_service('redis', redis_addr, 6379)
    srvcheck.wait_for_service('clickhouse', clickhouse_url, 8123)
    srvcheck.wait_for_service('planner', planner_url, 7997)

    logs.setup_logging('server')
    log.info('Kraken Server started, version %s', version.version)

    # Create the connexion application instance
    basedir = os.path.abspath(os.path.dirname(__file__))
    _, server_port = server_addr.split(':')
    connex_app = connexion.App('Kraken Server', port=int(server_port), specification_dir=basedir)

    # Get the underlying Flask app instance
    app = connex_app.app

    # Configure the SqlAlchemy part of the app instance
    app.config["SQLALCHEMY_ECHO"] = False
    app.config["SQLALCHEMY_DATABASE_URI"] = db_url + '?application_name=server'
    app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False

    # initialize SqlAlchemy
    models.db.init_app(app)

    with app.app_context():
        # setup sentry
        sentry_url = models.get_setting('monitoring', 'sentry_dsn')
        logs.setup_sentry(sentry_url)

        # check minio connection
        minio_conn = minioops.check_connection()
        if minio_conn:
            log.info('minio is up')
        else:
            minio_addr, _, _ = minioops.get_minio_addr()
            log.warning('No connection to minio at %s', minio_addr)

        # prepare access control
        access.init(redis_addr)

    # Read the swagger.yml file to configure the endpoints
    connex_app.add_api("swagger.yml", resolver=MyResolver())
    connex_app.add_api("swagger.yml", resolver=MyResolver(), base_path='/api')  # for backward compatibility

    # backend for serving agents
    connex_app.add_url_rule("/backend", view_func=backend.serve_agent_request, methods=['POST'])
    connex_app.add_url_rule("/bk/backend", view_func=backend.serve_agent_request, methods=['POST'])

    # serve agent files for agent update
    connex_app.add_url_rule("/install/<blob>", view_func=agentblob.serve_agent_blob, methods=['GET'])
    connex_app.add_url_rule("/bk/install/<blob>", view_func=agentblob.serve_agent_blob, methods=['GET'])

    # serve build artifacts
    connex_app.add_url_rule("/artifacts/<store_type>/f/<flow_id>/<path:path>", view_func=storage.serve_flow_artifact, methods=['GET'])
    connex_app.add_url_rule("/artifacts/<store_type>/r/<run_id>/<path:path>", view_func=storage.serve_run_artifact, methods=['GET'])
    connex_app.add_url_rule("/bk/artifacts/<store_type>/f/<flow_id>/<path:path>", view_func=storage.serve_flow_artifact, methods=['GET'])
    connex_app.add_url_rule("/bk/artifacts/<store_type>/r/<run_id>/<path:path>", view_func=storage.serve_run_artifact, methods=['GET'])

    # serve job log
    connex_app.add_url_rule("/job_log/<job_id>", view_func=job_log.serve_job_log, methods=['GET'])
    connex_app.add_url_rule("/bk/job_log/<job_id>", view_func=job_log.serve_job_log, methods=['GET'])
    connex_app.add_url_rule("/job_log/<job_id>/<step_idx>", view_func=job_log.serve_step_log, methods=['GET'])
    connex_app.add_url_rule("/bk/job_log/<job_id>/<step_idx>", view_func=job_log.serve_step_log, methods=['GET'])

    # install webhooks
    webhooks_bp = webhooks.create_blueprint()
    app.register_blueprint(webhooks_bp, url_prefix='/webhooks')
    app.register_blueprint(webhooks_bp, url_prefix='/bk/webhooks')

    # branch status badge
    app.add_url_rule("/branch-badge/<branch_id>", view_func=badge.get_branch_badge, methods=['GET'], defaults={'what': None})
    app.add_url_rule("/branch-badge/<branch_id>/<what>", view_func=badge.get_branch_badge, methods=['GET'])
    app.add_url_rule("/bk/branch-badge/<branch_id>", view_func=badge.get_branch_badge, methods=['GET'], defaults={'what': None})
    app.add_url_rule("/bk/branch-badge/<branch_id>/<what>", view_func=badge.get_branch_badge, methods=['GET'])

    # oidc/oauth2 redirect after login
    app.add_url_rule("/bk/oidc-logged", view_func=authn.oidc_logged, methods=['GET'])

    app.before_request(_set_log_ctx)
    app.teardown_request(_clear_request_ctx)

    app.register_error_handler(500, _unhandled_error_handler)

    return connex_app


def main():
    app = create_app()
    app.run(debug=True)


if __name__ == "__main__":
    main()



# Copyright 2021 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import base64
import logging
import datetime

import pytz

# KUBERNETES
import kubernetes

from .. import utils
from ..models import get_setting
from .common import _create_agent

log = logging.getLogger(__name__)


def _login(api_server_url):
    # Define the bearer token we are going to use to authenticate.
    # See here to create the token:
    # https://kubernetes.io/docs/tasks/access-application-cluster/access-cluster/
    token = get_setting('cloud', 'k8s_token')
    token = base64.b64decode(token).decode()

    # Create a configuration object
    cfg = kubernetes.client.Configuration()

    # Specify the endpoint of your Kube cluster
    cfg.host = api_server_url

    # Security part.
    # In this simple example we are not going to verify the SSL certificate of
    # the remote cluster (for simplicity reason)
    cfg.verify_ssl = False
    # Nevertheless if you want to do it you can with these 2 parameters
    # configuration.verify_ssl=True
    # ssl_ca_cert is the filepath to the file that contains the certificate.
    # configuration.ssl_ca_cert="certificate"

    cfg.api_key = {"authorization": "Bearer " + token}

    namespace = get_setting('cloud', 'k8s_namespace')

    # Create a ApiClient with our config
    api_client = kubernetes.client.ApiClient(cfg)
    return api_client, namespace


def _get_core_api():
    api_server_url = get_setting('cloud', 'k8s_api_server_url')
    if api_server_url:
        api_client, namespace = _login(api_server_url)
        core_api = kubernetes.client.CoreV1Api(api_client)
    else:
        kubernetes.config.load_incluster_config()
        core_api = kubernetes.client.CoreV1Api()
        namespace = get_setting('cloud', 'k8s_namespace')
    return core_api, namespace


def check_k8s_settings():
    namespace = get_setting('cloud', 'k8s_namespace')
    if not namespace:
        return 'Kubernetes namespace is empty'

    token = get_setting('cloud', 'k8s_token')
    api_server_url = get_setting('cloud', 'k8s_api_server_url')

    if api_server_url and not token:
        return 'Kubernetes token is empty while API server URL is provided'

    if api_server_url:
        api_client, _ = _login(api_server_url)
        core_api = kubernetes.client.CoreV1Api(api_client)
        ver_api = kubernetes.client.VersionApi(api_client)
    else:
        kubernetes.config.load_incluster_config()
        core_api = kubernetes.client.CoreV1Api()
        ver_api = kubernetes.client.VersionApi()

    try:
        res = ver_api.get_code()
        log.info('k8s api %s', res)

        res = core_api.read_namespace(namespace)
        log.info('k8s namespace %s', res)

        if res.status.phase != 'Active':
            return 'Namespace %s is not Active, it is %s' % (namespace, res.status.phase)
    except Exception as ex:
        log.exception('k8s get_code exc')
        return str(ex)

    return 'ok'


def create_pods(ag, system, num,
                server_url, clickhouse_addr):
    core_api, namespace = _get_core_api()

    # prepare create_container
    cmd = 'apt-get update && apt-get install -y --no-install-recommends ca-certificates sudo wget python3'
    cmd += ' && mkdir -p /opt/kraken'
    cmd += ' && wget -O /opt/kraken/kkagent {server_url}/bk/install/agent'
    cmd += ' && wget -O /opt/kraken/kktool {server_url}/bk/install/tool'
    cmd += ' && chmod a+x /opt/kraken/kkagent /opt/kraken/kktool'
    cmd += ' && mkdir -p /tmp/kk-jobs'
    cmd += ' && /opt/kraken/kkagent run -d /tmp/kk-jobs -s {server_url} -c {clickhouse_addr}'
    cmd += ' --system-id {system_id} --one-job'
    cmd = cmd.format(server_url=server_url, clickhouse_addr=clickhouse_addr,
                     system_id=system.id)

    container = kubernetes.client.V1Container(
        image=system.name,
        name='kraken-container',
        args=['args'],
        env=[{'name': 'LC_ALL', 'value': 'C.UTF-8'},
             {'name': 'LANG', 'value': 'C.UTF-8'}],
        command=['bash', '-c', cmd])

    log.info('Created container with name: %s, image: %s and args: %s',
             container.name, container.image, container.args)

    # prepare pod template
    # prepare job
    t0 = datetime.datetime(2020, 1, 1, tzinfo=pytz.utc)
    n = datetime.datetime.now(pytz.utc)
    dt = n - t0
    dt = int(dt.total_seconds())
    series = 'kk-%d-%d-%d' % (ag.id, system.id, dt)

    triggered_pods = []
    for i in range(num):
        pod_name = '%s-%d' % (series, i)
        pod_def = kubernetes.client.V1Pod(
            spec=kubernetes.client.V1PodSpec(restart_policy="Never", containers=[container]),
            metadata=kubernetes.client.V1ObjectMeta(name=pod_name, labels={'series': series,
                                                                           'group': str(ag.id),
                                                                           'name': pod_name}))
        # start pod
        res = core_api.create_namespaced_pod(namespace, pod_def)
        triggered_pods.append(res)
        #print('POD ', res)

    log.info('group: %s, sys: %s: started k8s %d pods for %s series',
             ag.name, system.name, num, series)

    while True:
        pods = core_api.list_namespaced_pod(namespace, label_selector="series=%s" % series, pretty=True)
        #print('pods ', pods)

        # check if all started otherwise wait for pending
        all_started = True
        for pod in pods.items:
            if pod.status.phase == 'Pending':
                all_started = False
        if all_started:
            break

        time.sleep(1)
        continue

    pods = core_api.list_namespaced_pod(namespace, label_selector="series=%s" % series, pretty=True)
    for pod in pods.items:
        #print('pod ', pod.metadata.name, pod.status)
        if pod.status.phase != 'Running':
            log.error('unexpected state of pod %s: %s', pod.metadata.name, pod.status.phase)
            continue

        params = dict(name=pod.metadata.name,
                      address=pod.status.pod_ip,
                      ip_address=pod.status.host_ip,
                      extra_attrs=dict(system=system.id,
                                       series=series,
                                       instance_id=pod.metadata.name,
                                       namespace=namespace))
        a = _create_agent(params, ag)
        log.info('spawned new agent %s in K8S pod %s', a, pod.metadata.name)
        #print('PARAMS ', params)


def destroy_pod(ag, agent):  # pylint: disable=unused-argument
    namespace = agent.extra_attrs['namespace']
    pod_name = agent.extra_attrs['instance_id']

    core_api, _ = _get_core_api()

    log.info('delete K8S pod %s', pod_name)
    try:
        core_api.delete_namespaced_pod(pod_name, namespace)
    except Exception:
        log.exception('IGNORED EXCEPTION')


def pod_exists(ag, agent):  # pylint: disable=unused-argument
    namespace = agent.extra_attrs['namespace']
    pod_name = agent.extra_attrs['instance_id']

    core_api, _ = _get_core_api()

    log.info('does K8S pod %s exist', pod_name)
    try:
        core_api.read_namespaced_pod(pod_name, namespace)
    except Exception:
        log.exception('IGNORED EXCEPTION')
        return False

    return True


def cleanup_dangling_pods(ag):
    core_api, namespace = _get_core_api()

    now = utils.utcnow()

    instances = 0
    terminated_instances = 0
    assigned_instances = 0
    orphaned_instances = 0
    orphaned_terminated_instances = 0

    pods = core_api.list_namespaced_pod(namespace, label_selector="group=%s" % ag.id, pretty=True)
    for pod in pods.items:

        # if assigned to some agent then skip it
        assigned = False
        for aa in ag.agents:
            agent = aa.agent
            if agent.extra_attrs and 'instance_id' in agent.extra_attrs and agent.extra_attrs['instance_id'] == pod.metadata.name:
                assigned = True
                break
        if assigned:
            assigned_instances += 1
            continue

        # instances have to be old enough to avoid race condition with
        # case when instances are being created but not yet assigned to agents
        created_at = pod.status.start_time
        if not created_at or now - created_at < datetime.timedelta(minutes=10):
            continue

        # the instance is not terminated, not assigned, old enough
        # so delete it as it seems to be a lost instance
        log.info('terminating lost K8S pod %s', pod.metadata.name)
        orphaned_instances += 1
        try:
            core_api.delete_namespaced_pod(pod.metadata.name, namespace)
        except Exception:
            log.exception('IGNORED EXCEPTION')

        orphaned_terminated_instances += 1

    return instances, terminated_instances, assigned_instances, orphaned_instances, orphaned_terminated_instances



# Copyright 2021 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from .. import consts
from . import aws_ec2
from . import aws_ecs
from . import azure
from . import k8s


log = logging.getLogger(__name__)


def create_machines(ag, system, num,
                    server_url, clickhouse_addr):
    method, _ = ag.get_deployment()
    if method == consts.AGENT_DEPLOYMENT_METHOD_AWS_EC2:
        aws_ec2.create_vms(ag, system, num,
                           server_url, clickhouse_addr)

    elif method == consts.AGENT_DEPLOYMENT_METHOD_AWS_ECS_FARGATE:
        aws_ecs.create_fargate_tasks(ag, system, num,
                                     server_url, clickhouse_addr)

    elif method == consts.AGENT_DEPLOYMENT_METHOD_AZURE_VM:
        azure.create_vms(ag, system, num,
                         server_url, clickhouse_addr)

    elif method == consts.AGENT_DEPLOYMENT_METHOD_K8S:
        k8s.create_pods(ag, system, num,
                        server_url, clickhouse_addr)

    else:
        raise NotImplementedError('deployment method %s not supported' % method)


def destroy_machine(ag, agent):
    method, _ = ag.get_deployment()
    if method == consts.AGENT_DEPLOYMENT_METHOD_AWS_EC2:
        aws_ec2.destroy_vm(ag, agent)
    elif method == consts.AGENT_DEPLOYMENT_METHOD_AWS_ECS_FARGATE:
        aws_ecs.destroy_fargate_task(ag, agent)
    elif method == consts.AGENT_DEPLOYMENT_METHOD_AZURE_VM:
        azure.destroy_vm(ag, agent)
    elif method == consts.AGENT_DEPLOYMENT_METHOD_K8S:
        k8s.destroy_pod(ag, agent)

    else:
        raise NotImplementedError('deployment method %s not supported' % method)


def check_if_machine_exists(ag, agent):
    method, _ = ag.get_deployment()
    if method == consts.AGENT_DEPLOYMENT_METHOD_AWS_EC2:
        return aws_ec2.vm_exists(ag, agent)
    if method == consts.AGENT_DEPLOYMENT_METHOD_AZURE_VM:
        return azure.vm_exists(ag, agent)
    if method == consts.AGENT_DEPLOYMENT_METHOD_K8S:
        return k8s.pod_exists(ag, agent)
    raise NotImplementedError('deployment method %s not supported' % method)


def cleanup_dangling_machines(ag):
    method, _ = ag.get_deployment()
    if method == consts.AGENT_DEPLOYMENT_METHOD_AWS_EC2:
        return aws_ec2.cleanup_dangling_vms(ag)
    if method == consts.AGENT_DEPLOYMENT_METHOD_AZURE_VM:
        return azure.cleanup_dangling_vms(ag)
    if method == consts.AGENT_DEPLOYMENT_METHOD_K8S:
        return k8s.cleanup_dangling_pods(ag)
    return 0, 0, 0, 0, 0


#if __name__ == '__main__':
#    sys_img = 'ubuntu:20.04'
#    #sys_img = 'public.ecr.aws/kraken-ci/kkagent:0.627'
#    main(sys_img, 'http://89.65.138.0:8080', '89.65.138.0:6363', '89.65.138.0:9999', '89.65.138.0:9001')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kraken',
 'kraken.migrations',
 'kraken.migrations.versions',
 'kraken.server',
 'kraken.server.bg',
 'kraken.server.cloud']

package_data = \
{'': ['*']}

install_requires = \
['Authlib>=1.1.0,<2.0.0',
 'Flask-SQLAlchemy>=2.5.1,<3.0.0',
 'Flask>=1.1.4,<2.0.0',
 'MarkupSafe==2.0.1',
 'RestrictedPython==5.0',
 'SQLAlchemy>=1.4.25,<2.0.0',
 'alembic>=1.7.3,<2.0.0',
 'apscheduler>=3.8.0,<4.0.0',
 'azure-identity>=1.6.1,<2.0.0',
 'azure-mgmt-compute>=23.0.0,<24.0.0',
 'azure-mgmt-monitor>=2.0.0,<3.0.0',
 'azure-mgmt-network>=19.0.0,<20.0.0',
 'azure-mgmt-resource>=19.0.0,<20.0.0',
 'azure-mgmt-storage>=18.0.0,<19.0.0',
 'azure-mgmt-subscription>=1.0.0,<2.0.0',
 'boto3>=1.18.52,<2.0.0',
 'casbin>=1.17.1,<2.0.0',
 'clickhouse-driver>=0.2.5,<0.3.0',
 'connexion>=2.13.1,<3.0.0',
 'furl>=2.1.3,<3.0.0',
 'giturlparse>=0.10.0,<0.11.0',
 'gunicorn>=20.1.0,<21.0.0',
 'jsonschema>=4.5.0,<5.0.0',
 'kubernetes>=20.13.0,<21.0.0',
 'minio>=7.1.0,<8.0.0',
 'passlib>=1.7.4,<2.0.0',
 'psycopg2-binary>=2.9.1,<3.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'python-ldap>=3.4.3,<4.0.0',
 'pytimeparse>=1.1.8,<2.0.0',
 'redis>=3.5.3,<4.0.0',
 'requests>=2.26.0,<3.0.0',
 'rq>=1.10.0,<2.0.0',
 'sentry-sdk[flask]>=1.5.0,<2.0.0',
 'setuptools>=66.0.0,<67.0.0',
 'swagger-ui-bundle>=0.0.9,<0.0.10',
 'tzlocal==2.1']

entry_points = \
{'console_scripts': ['kkdbmigrate = kraken.migrations.apply:main',
                     'kkplanner = kraken.server.planner:main',
                     'kkqneck = kraken.server.qneck:main',
                     'kkrq = kraken.server.kkrq:main',
                     'kkscheduler = kraken.server.scheduler:main',
                     'kkwatchdog = kraken.server.watchdog:main']}

setup_kwargs = {
    'name': 'krakenci-server',
    'version': '1.0',
    'description': 'Kraken CI server.',
    'long_description': '# Kraken CI\n\n![Release Version](https://img.shields.io/github/v/release/Kraken-CI/kraken)\n![Release Date](https://img.shields.io/github/release-date/Kraken-CI/kraken)\n\n![Kraken Build](https://lab.kraken.ci/bk/branch-badge/2)\n![Kraken Tests](https://lab.kraken.ci/bk/branch-badge/2/tests)\n![Kraken Issues](https://lab.kraken.ci/bk/branch-badge/2/issues)\n\n[![Codacy Badge](https://app.codacy.com/project/badge/Grade/be770bd29e374ece9e6f2782a1de99fc)](https://www.codacy.com/gh/Kraken-CI/kraken/dashboard?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=Kraken-CI/kraken&amp;utm_campaign=Badge_Grade)\n[![Language grade: Python](https://img.shields.io/lgtm/grade/python/g/Kraken-CI/kraken.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/Kraken-CI/kraken/context:python)\n[![Language grade: JavaScript](https://img.shields.io/lgtm/grade/javascript/g/Kraken-CI/kraken.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/Kraken-CI/kraken/context:javascript)\n[![Total alerts](https://img.shields.io/lgtm/alerts/g/Kraken-CI/kraken.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/Kraken-CI/kraken/alerts/)\n[![codebeat badge](https://codebeat.co/badges/556ac028-2390-4093-839e-a509f5678cf1)](https://codebeat.co/projects/github-com-kraken-ci-kraken-master)\n\n[![CII Best Practices](https://bestpractices.coreinfrastructure.org/projects/4758/badge)](https://bestpractices.coreinfrastructure.org/projects/4758)\n[![Artifact Hub](https://img.shields.io/endpoint?url=https://artifacthub.io/badge/repository/kraken-ci)](https://artifacthub.io/packages/search?repo=kraken-ci)\n\n\n<!-- ABOUT THE PROJECT -->\n## About Kraken CI\n\n![Kraken CI Results Page](https://kraken.ci/img/slide-branch-results.png)\n\nKraken CI is a modern, open-source, on-premise CI/CD system\nthat is highly scalable and focused on testing.\n\nMore information can be found on https://kraken.ci\n\n\n<!-- GETTING STARTED -->\n## Getting Started\n\nQuick start guide is here: https://kraken.ci/docs/quick-start\n\nFull installation manual: https://kraken.ci/docs/installation\n\nAnd here is developers guide: https://kraken.ci/docs/dev-guide\n\n\n<!-- USAGE EXAMPLES -->\n## Usage\n\nGuides can be found here: https://kraken.ci/docs/guide-intro\n\nDemo site is available here: https://lab.kraken.ci/\n\n\n<!-- ROADMAP -->\n## Roadmap\n\nSee the [open issues](https://github.com/kraken-ci/kraken/issues) for a list of proposed features (and known issues).\n\n\n<!-- CONTRIBUTING -->\n## Contributing\n\nContributions are what make the open source community such an amazing place to be learn, inspire, and create. Any contributions you make are **greatly appreciated**.\n\nDetails on https://kraken.ci/docs/contrib-kraken\n\n\n<!-- LICENSE -->\n## License\n\nDistributed under the Apache 2.0 License. See `LICENSE` for more information.\n\n\n<!-- CONTACT -->\n## Contact\n\nMichal Nowikowski - godfryd@gmail.com\n\nProject Link: [https://kraken.ci](https://kraken.ci)\n',
    'author': 'Michal Nowikowski',
    'author_email': 'godfryd@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://kraken.ci/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# Copyright 2020-2021 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import time
import logging
import xmlrpc.client
from collections import defaultdict

from flask import Flask
from sqlalchemy.sql.expression import asc, desc
from sqlalchemy.orm.attributes import flag_modified
from sqlalchemy.orm import joinedload
import giturlparse
import pytimeparse
import redis

from ..models import db, Run, Job, TestCaseResult, Branch, Flow, Stage, Project, get_setting
from ..models import AgentsGroup, Agent, System, TestCaseComment, Tool
from ..models import RepoChanges, Secret
from ..schema import prepare_new_planner_triggers
from ..schema import check_and_correct_stage_schema
from ..cloud import cloud
from .. import exec_utils  # pylint: disable=cyclic-import
from .. import consts
from .. import notify
from .. import logs
from .. import gitops
from .. import kkrq
from .. import utils
from .. import dbutils
from .. import toolops
from .. import toolutils


log = logging.getLogger(__name__)


def _create_app(task_name):
    # addresses
    db_url = os.environ.get('KRAKEN_DB_URL', consts.DEFAULT_DB_URL)

    logs.setup_logging('rq')
    log.set_ctx(tool=task_name)

    # Create  Flask app instance
    app = Flask('Kraken Background')

    # Configure the SqlAlchemy part of the app instance
    app.config["SQLALCHEMY_ECHO"] = False
    app.config["SQLALCHEMY_DATABASE_URI"] = db_url + '?application_name=rq_' + task_name
    app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False

    # initialize SqlAlchemy
    db.init_app(app)

    # setup sentry
    with app.app_context():
        sentry_url = get_setting('monitoring', 'sentry_dsn')
        logs.setup_sentry(sentry_url)

    return app


def _trigger_stages(run):
    """Trigger the following stages after just completed run_id stage."""
    log.info('starting triggering stages after run %s', run)

    # go through next stages and trigger them if needed
    curr_stage_name = run.stage.name
    branch = run.stage.branch
    for stage in branch.stages:
        if stage.deleted:
            continue
        if stage.schema['parent'] != curr_stage_name:
            # skip stages that are not childs of completed run's stage
            continue
        if not stage.schema['triggers'].get('parent', False) and not stage.schema['triggers'].get('manual', False):
            # skip stages that have parent trigger set to False
            # or that have no manual trigger
            continue

        if stage.enabled:
            exec_utils.start_run(stage, run.flow, reason=dict(reason='parent', run_id=run.id))
        else:
            log.info('stage %s not started because it is disabled', stage)


def _prepare_flow_summary(flow):
    if flow.kind != consts.FLOW_KIND_CI:
        return

    # update pointer to last, completed, CI flow in the branch
    # and if needed then reset pointer to incomplete flow

    # update completed
    last_flow = flow.branch.ci_last_completed_flow
    if last_flow is None or last_flow.created < flow.created:
        flow.branch.ci_last_completed_flow = flow

    # update incomplete
    last_flow = flow.branch.ci_last_incomplete_flow
    if last_flow and last_flow.id == flow.id:
        flow.branch.ci_last_incomplete_flow = None

    db.session.commit()

    # get flow summary and store it in redis to cache it;
    # it will be used e.g. to provide badge info

    # get redis reference
    redis_addr = os.environ.get('KRAKEN_REDIS_ADDR', consts.DEFAULT_REDIS_ADDR)
    redis_host, redis_port = utils.split_host_port(redis_addr, 6379)
    rds = redis.Redis(host=redis_host, port=redis_port, db=consts.REDIS_KRAKEN_DB)

    # prepare flow summary
    errors = False
    tests_passed = 0
    tests_total = 0
    tests_regr = 0
    tests_fix = 0
    issues_total = 0
    issues_new = 0
    for r in flow.runs:
        if r.jobs_error > 0:
            errors = True
        tests_passed += r.tests_passed
        tests_total += r.tests_total
        tests_regr += r.regr_cnt
        tests_fix += r.fix_cnt
        issues_total += r.issues_total
        issues_new += r.issues_new
    val = dict(id=flow.id,
               label=flow.get_label(),
               errors=errors,
               tests_passed=tests_passed,
               tests_total=tests_total,
               tests_regr=tests_regr,
               tests_fix=tests_fix,
               issues_total=issues_total,
               issues_new=issues_new)

    # store in redis flow state for badge and cctray
    key = 'branch-%d' % flow.branch_id
    val_ext = val.copy()
    val_ext['project'] = flow.branch.project.name
    val_ext['branch'] = flow.branch.name
    val_ext['activity'] = 'Sleeping'
    val_ext['lastBuildTime'] = utils.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ")
    val_ext['url'] = '%s/branches/%d/ci' % (get_setting('general', 'server_url'), flow.branch_id)
    rds.set(key, json.dumps(val_ext))
    log.info('cached flow results: %s = %s', key, val_ext)

    # store in db as well, it will be used for charts in UI,
    flow.summary = val
    db.session.commit()


def analyze_run(run_id):
    log.reset_ctx()
    app = _create_app('analyze_run_%d' % run_id)

    with app.app_context():
        run = Run.query.filter_by(id=run_id).one_or_none()
        if run is None:
            log.error('got unknown run to analyze: %s', run_id)
            return

        log.set_ctx(branch=run.flow.branch_id, flow_kind=run.flow.kind, flow=run.flow_id, run=run.id)
        log.info('starting analysis of run %s', run)

        # calculate run stats
        run.tests_total = run.tests_passed = run.tests_not_run = 0
        run.jobs_error = run.jobs_total = 0
        run.issues_total = 0
        non_covered_jobs = Job.query.filter_by(run=run).filter_by(covered=False).all()
        for job in non_covered_jobs:
            # calculate job tests stats
            for job_tcr in job.results:
                run.tests_total += 1
                if job_tcr.result == consts.TC_RESULT_PASSED:
                    run.tests_passed += 1
                elif job_tcr.result == consts.TC_RESULT_NOT_RUN:
                    run.tests_not_run += 1

            # calculate issues stats
            run.issues_total += len(job.issues)

            # calculate jobs stats
            run.jobs_total += 1
            if job.completion_status not in [consts.JOB_CMPLT_ALL_OK, None]:
                run.jobs_error += 1
            db.session.commit()

        # trigger any following stages to currently completed run if there was no errors
        # and there were some jobs executed (sometimes jobs are not started at all
        # due to some internal problems)
        if run.jobs_error == 0 and run.jobs_total > 0:
            _trigger_stages(run)

        # establish new state for flow
        flow = run.flow
        is_completed = True
        log.info('check if flow %s is completed', flow)
        for r in flow.runs:
            log.info('  run %s: state: %d', r, r.state)
            if r.state == consts.RUN_STATE_IN_PROGRESS:
                is_completed = False
                break
        log.info('flow %s: %s completed', flow, 'IS' if is_completed else 'NOT')

        if is_completed:
            log.info('completed flow %s', flow)
            _prepare_flow_summary(flow)

            now = utils.utcnow()
            flow.finished = now
            flow.state = consts.FLOW_STATE_COMPLETED
            db.session.commit()

        kkrq.enq(analyze_results_history, run.id)
        log.info('run %s analysis completed, started analyze results history', run)


def _analyze_ci_test_case_result(job, job_tcr):
    # log.info('TCR: %s %s %s %s', job_tcr, job_tcr.test_case.name, job_tcr.job.run.flow, job_tcr.job.run.flow.created)

    # get previous 10 results
    q = TestCaseResult.query
    q = q.options(joinedload('comment'))
    q = q.filter(TestCaseResult.id != job_tcr.id)
    q = q.filter_by(test_case_id=job_tcr.test_case_id)
    q = q.join('job')
    q = q.filter_by(covered=False)
    q = q.filter_by(agents_group=job.agents_group)
    q = q.filter_by(system=job.system)
    q = q.join('job', 'run', 'flow', 'branch')
    q = q.filter(Branch.id == job.run.flow.branch_id)
    q = q.filter(Flow.kind == 0)  # CI
    q = q.filter(Flow.created < job.run.flow.created)
    q = q.order_by(desc(Flow.created))
    q = q.limit(10)

    # import sqlalchemy.dialects.postgresql
    # s = q.statement.compile(compile_kwargs={"literal_binds": True},
    #                         dialect=sqlalchemy.dialects.postgresql.dialect())
    # log.info(' query: %s', s)

    tcrs = q.all()
    tcrs.reverse() # sort from oldest to latest
    # determine instability and find last test case comment
    tcc = None
    for idx, tcr in enumerate(tcrs):
        # log.info('TCR: %s %s %s', tcr, tcr.test_case.name, tcr.job.run.flow.created)
        if idx == 0:
            job_tcr.instability = 0
        elif tcr.result != tcrs[idx - 1].result:
            job_tcr.instability += 1

        if tcr.comment:
            tcc = tcr.comment

    # if there is no comment in current tcr and there was some in the past, and current result is not passed
    # then assing it to current one
    if not job_tcr.comment and job_tcr.result != consts.TC_RESULT_PASSED:
        if tcc is None:
            q = TestCaseComment.query
            q = q.filter_by(test_case=job_tcr.test_case, branch=job_tcr.job.run.flow.branch)
            tcc = q.one_or_none()
            if tcc:
                # comment was not present in last 10 flows so move it to new state
                tcc.state = consts.TC_COMMENT_NEW
        job_tcr.comment = tcc

        # update last flow in tcc if needed
        if tcc and tcc.last_flow.created < job_tcr.job.run.flow.created:
            tcc.last_flow = job_tcr.job.run.flow

    # determine age and change
    if len(tcrs) > 0:
        prev_tcr = tcrs[-1]
        # log.info('PREV TCR: %s %s %s %s', prev_tcr, prev_tcr.test_case.name, prev_tcr.job.run.flow, prev_tcr.job.run.flow.created)
        if prev_tcr.result == job_tcr.result:
            job_tcr.age = prev_tcr.age + 1
            job_tcr.change = consts.TC_RESULT_CHANGE_NO
            # log.info('PREV TCR: %s no change', prev_tcr)
        else:
            job_tcr.instability += 1
            job_tcr.age = 0
            if job_tcr.result == consts.TC_RESULT_PASSED and prev_tcr.result != consts.TC_RESULT_PASSED:
                job_tcr.change = consts.TC_RESULT_CHANGE_FIX
                # log.info('PREV TCR: %s fix', prev_tcr)
            elif job_tcr.result != consts.TC_RESULT_PASSED and prev_tcr.result == consts.TC_RESULT_PASSED:
                job_tcr.change = consts.TC_RESULT_CHANGE_REGR
                # log.info('PREV TCR: %s regr', prev_tcr)
            # else:
            #     log.info('PREV TCR: %s %s vs %s', prev_tcr, job_tcr.result, prev_tcr.result)
    else:
        # log.info('NO PREV TCR')
        job_tcr.change = consts.TC_RESULT_CHANGE_NEW

    # determine relevancy
    # 0 initial
    job_tcr.relevancy = 0
    # +1 not pass
    if job_tcr.result != consts.TC_RESULT_PASSED:
        job_tcr.relevancy += 1
        # +1 no comment or not root caused comment state
        if not job_tcr.comment or job_tcr.comment.state not in [consts.TC_COMMENT_BUG_IN_PRODUCT, consts.TC_COMMENT_BUG_IN_TEST]:
            job_tcr.relevancy += 1
    # +1 failure
    if job_tcr.result == consts.TC_RESULT_FAILED:
        job_tcr.relevancy += 1
    # +1 instability <= 3
    if job_tcr.instability <= 3:
        job_tcr.relevancy += 1
    # +1 age < 5
    if job_tcr.age < 5:
        job_tcr.relevancy += 1
    # +1 regression (age=0)
    if job_tcr.change == consts.TC_RESULT_CHANGE_REGR:
        job_tcr.relevancy += 1


def _analyze_dev_test_case_result(job, job_tcr):
    # get reference result from CI flow
    q = TestCaseResult.query
    q = q.filter_by(test_case_id=job_tcr.test_case_id)
    q = q.join('job')
    q = q.filter_by(covered=False)
    q = q.filter_by(agents_group=job_tcr.job.agents_group)
    q = q.filter_by(system=job.system)
    q = q.filter_by(state=consts.JOB_STATE_COMPLETED)
    q = q.join('job', 'run', 'flow', 'branch')
    q = q.filter(Branch.id == job.run.flow.branch_id)
    q = q.filter(Flow.kind == consts.FLOW_KIND_CI)
    q = q.order_by(desc(Flow.created))

    ref_tcr = q.first()

    # determine change
    if ref_tcr:
        # log.info('REF TCR: %s %s %s %s', ref_tcr, ref_tcr.test_case.name, ref_tcr.job.run.flow, ref_tcr.job.run.flow.created)
        # copy ref tcr stats to current dev results so user can see how it behaves
        job_tcr.age = ref_tcr.age
        job_tcr.instability = ref_tcr.instability

        if ref_tcr.result == job_tcr.result:
            job_tcr.change = consts.TC_RESULT_CHANGE_NO
        else:
            if job_tcr.result == consts.TC_RESULT_PASSED and ref_tcr.result != consts.TC_RESULT_PASSED:
                job_tcr.change = consts.TC_RESULT_CHANGE_FIX
            elif job_tcr.result != consts.TC_RESULT_PASSED and ref_tcr.result == consts.TC_RESULT_PASSED:
                job_tcr.change = consts.TC_RESULT_CHANGE_REGR
    else:
        log.info('NO REF TCR, ie. new')
        job_tcr.change = consts.TC_RESULT_CHANGE_NEW


def _analyze_job_results_history(job):
    # TODO: if branch is forked from another base branch take base branch results into account

    q = TestCaseResult.query
    q = q.filter_by(job=job)

    counts = [0, 0, 0, 0]
    cnt = 0
    total = 0
    for job_tcr in q.all():
        t0 = time.time()
        # analyze history
        if job.run.flow.kind == consts.FLOW_KIND_CI:  # CI
            _analyze_ci_test_case_result(job, job_tcr)
        else:  # DEV
            _analyze_dev_test_case_result(job, job_tcr)

        db.session.commit()

        t1 = time.time()
        total += t1 - t0
        cnt += 1
        log.info('Analyzed result dt:%0.5f %s %s: %s', t1 - t0, job_tcr, job_tcr.test_case.name,
                 consts.TC_RESULT_CHANGES_NAME[job_tcr.change])

        counts[job_tcr.change] += 1

    # current avg time is 0.0295s,
    log.info('Analyzed in time:%0.5f, avg:%0.5f, count:%d',
             total, total / cnt if cnt > 0 else 0, cnt)

    return counts


def _analyze_job_issues_history(job):
    q = Job.query
    q = q.filter_by(name=job.name)
    q = q.filter_by(completion_status=consts.JOB_CMPLT_ALL_OK)
    q = q.filter_by(system=job.system)
    q = q.filter_by(agents_group=job.agents_group)
    q = q.join('run')
    q = q.filter_by(stage=job.run.stage)
    q = q.join('run', 'flow')
    q = q.filter(Flow.created < job.run.flow.created)
    q = q.filter(Flow.kind == 0)  # CI
    q = q.order_by(desc(Flow.created))
    prev_job = q.first()
    if prev_job is None:
        log.info('prev job for issues not found')
        return 0

    issues_new = 0

    for issue in job.issues:
        log.info('Analyze issue %s', issue)
        prev_issue = None
        dist = 1000
        log.info('PREV ISSES %s', len(prev_job.issues))
        for i in prev_job.issues:
            if i.path != issue.path or i.issue_type != issue.issue_type or i.symbol !=issue.symbol:
                continue
            new_dist = abs(i.line - issue.line)
            if new_dist > 5:
                continue
            log.info('issue %s - prev issue %s, dist %s', issue, i, new_dist)
            if new_dist < dist:
                dist = new_dist
                prev_issue = i

        if prev_issue is not None:
            issue.age = prev_issue.age + 1
        if issue.age == 0:
            issues_new += 1
    db.session.commit()

    return issues_new


def _analyze_results_history(run_id):
    run = Run.query.filter_by(id=run_id).one_or_none()
    if run is None:
        log.error('got unknown run to analyze results history: %s', run_id)
        return

    log.set_ctx(branch=run.flow.branch_id, flow_kind=run.flow.kind, flow=run.flow_id, run=run.id)

    log.info('starting results history analysis of run %s, flow %s [%s] ',
             run, run.flow, 'CI' if run.flow.kind == 0 else 'DEV')

    # check prev run in case of CI
    if run.flow.kind == 0:
        q = Run.query
        q = q.filter_by(deleted=None)
        q = q.filter_by(stage_id=run.stage_id)
        q = q.join('flow')
        q = q.filter(Flow.created < run.flow.created)
        q = q.filter(Flow.kind == run.flow.kind)
        q = q.order_by(desc(Flow.created))
        prev_run = q.first()
        if prev_run is None:
            log.info('this is the first run %s in the first flow %s of branch %s',
                     run, run.flow, run.flow.branch)

        elif prev_run.state not in [consts.RUN_STATE_PROCESSED, consts.RUN_STATE_MANUAL]:
            # prev run is not processed yet
            log.info('postpone anlysis of run %s as prev run %s is not processed yet', run, prev_run)
            return

    # analyze jobs history of this run
    run.new_cnt = run.no_change_cnt = run.regr_cnt = run.fix_cnt = 0
    run.issues_new = 0
    non_covered_jobs = Job.query.filter_by(run=run).filter_by(covered=False).all()
    for job in non_covered_jobs:
        log.set_ctx(job=job.id)
        # analyze results history
        counts = _analyze_job_results_history(job)
        no_change_cnt, fix_cnt, regr_cnt, new_cnt = counts
        run.no_change_cnt += no_change_cnt
        run.fix_cnt += fix_cnt
        run.regr_cnt += regr_cnt
        run.new_cnt += new_cnt
        db.session.commit()

        # analyze issues history
        issues_new = _analyze_job_issues_history(job)
        run.issues_new += issues_new
        db.session.commit()
        # log.info('job %s: new:%d no-change:%d regr:%d fix:%d',
        #          job, new_cnt, no_change_cnt, regr_cnt, fix_cnt)
    log.set_ctx(job=None)

    log.info('run %s: new:%d no-change:%d regr:%d fix:%d',
             run, run.new_cnt, run.no_change_cnt, run.regr_cnt, run.fix_cnt)

    run.state = consts.RUN_STATE_PROCESSED
    run.processed_at = utils.utcnow()
    db.session.commit()
    log.info('history anlysis of run %s completed', run)

    kkrq.enq(notify_about_completed_run, run.id)

    # trigger analysis of the following run
    q = Run.query
    q = q.filter_by(stage_id=run.stage_id)
    q = q.join('flow')
    q = q.filter(Flow.created > run.flow.created)
    q = q.filter(Flow.kind == run.flow.kind)
    q = q.order_by(asc(Flow.created))
    next_run = q.first()
    if next_run is not None and next_run.state in [consts.RUN_STATE_COMPLETED, consts.RUN_STATE_PROCESSED]:
        kkrq.enq_neck(analyze_results_history, next_run.id)

    log.info('finished results history analysis of run %s, flow %s [%s]',
             run, run.flow, 'CI' if run.flow.kind == 0 else 'DEV')


def analyze_results_history(run_id):
    log.reset_ctx()
    app = _create_app('analyze_results_history_%d' % run_id)

    with app.app_context():
        _analyze_results_history(run_id)


def notify_about_started_run(run_id):
    log.reset_ctx()
    app = _create_app('notify_about_started_run_%d' % run_id)

    with app.app_context():
        run = Run.query.filter_by(id=run_id).one_or_none()
        if run is None:
            log.error('got unknown run to notify: %s', run_id)
            return

        log.set_ctx(branch=run.flow.branch_id, flow_kind=run.flow.kind, flow=run.flow_id, run=run.id)
        log.info('starting notification about started run %s', run)

        notify.notify(run, 'start')


def notify_about_completed_run(run_id):
    log.reset_ctx()
    app = _create_app('notify_about_completed_run_%s' % run_id)

    with app.app_context():
        run = Run.query.filter_by(id=run_id).one_or_none()
        if run is None:
            log.error('got unknown run to notify: %s', run_id)
            return

        log.set_ctx(branch=run.flow.branch_id, flow_kind=run.flow.kind, flow=run.flow_id, run=run.id)
        log.info('starting notification about completed run %s', run)

        notify.notify(run, 'end')


def _estimate_timeout(job):
    if not job.assigned or not job.finished:
        log.warning('job %s has None dates: %s %s', job, job.assigned, job.finished)
        return
    if job.assigned > job.finished:
        log.warning('job %s has wrong dates: %s %s', job, job.assigned, job.finished)
        return

    q = Job.query
    q = q.filter_by(name=job.name)
    q = q.filter_by(state=consts.JOB_STATE_COMPLETED)
    q = q.filter(Job.completion_status.in_([consts.JOB_CMPLT_ALL_OK, consts.JOB_CMPLT_JOB_TIMEOUT]))
    q = q.filter_by(covered=False)
    q = q.filter_by(agents_group=job.agents_group)
    q = q.filter_by(system=job.system)
    # TODO add filtering by config
    q = q.join('run')
    q = q.filter(Run.stage_id == job.run.stage_id)
    q = q.join('run', 'flow')
    q = q.filter(Flow.created <= job.run.flow.created)
    q = q.order_by(desc(Flow.created))
    q = q.limit(10)

    jobs = q.all()
    if len(jobs) < 4:
        log.info('not enough jobs to estimate timeout')
        return

    # check job timestamps
    if not job.assigned:
        log.info('job %s assigned in None', job)
    if not job.finished:
        log.info('job %s finished in None', job)
    if job.assigned > job.finished:
        log.info('job %s assigned %s > finished %s', job, job.assigned, job.finished)
    if not job.assigned or not job.finished or job.assigned > job.finished:
        max_duration = duration = consts.DEFAULT_JOB_TIMEOUT
    else:
        max_duration = duration = job.finished - job.assigned

    # find max duration in last 10 jobs
    timeout_occured = False
    prev_job_id = job.id
    for j in jobs:
        if not j.assigned or not j.finished:
            log.warning('job %s has None dates: %s %s', j, j.assigned, j.finished)
            continue
        if j.assigned > j.finished:
            log.warning('job %s has wrong dates: %s %s', j, j.assigned, j.finished)
            continue
        if j.completion_status == consts.JOB_CMPLT_JOB_TIMEOUT:
            timeout_occured = True

        if j.id == prev_job_id:
            # cumulate duration if this is the same job (TODO: why?)
            duration += j.finished - j.assigned
        else:
            duration = j.finished - j.assigned
            prev_job_id = j.id

        if duration > max_duration:
            max_duration = duration

    # estimate new timeout from max duration
    timeout = int(max_duration.total_seconds() * 1.7)
    timeout = max(timeout, 60)

    stage = job.run.stage
    job_key = "%s-%d-%d" % (job.name, job.system_id, job.agents_group_id)

    if stage.timeouts is None:
        stage.timeouts = {}

    # if timeout occured in last 10 jobs than make new timeout adjustment
    if timeout_occured:
        old_timeout = stage.timeouts.get(job_key, consts.DEFAULT_JOB_TIMEOUT)
        log.info('new: %s, old: %s', timeout, old_timeout)
        if timeout < old_timeout:
            timeout = old_timeout * 2

        for j in stage.schema['jobs']:
            if j['name'] == job.name:
                user_timeout = j.get('timeout', consts.DEFAULT_JOB_TIMEOUT)
                break

        log.info('new: %s, user: %s', timeout, user_timeout)
        timeout = max(timeout, user_timeout)

    # it does not make sense to make timeout bigger than 3 days ie. 259200 seconds
    timeout = max(timeout, 259200)

    log.info("new timeout for job '%s' in stage '%s': %ssecs", job_key, stage.name, timeout)
    stage.timeouts[job_key] = timeout
    flag_modified(stage, 'timeouts')
    db.session.commit()


def job_completed(job_id):
    log.reset_ctx()
    app = _create_app('job_completed_%d' % job_id)

    with app.app_context():

        now = utils.utcnow()

        job = Job.query.filter_by(id=job_id).one_or_none()
        if job is None:
            log.error('got unknown job: %s', job_id)
            return

        log.set_ctx(branch=job.run.flow.branch_id, flow_kind=job.run.flow.kind, flow=job.run.flow_id, run=job.run_id, job=job.id)
        log.info('completing job %s', job)

        if job.state != consts.JOB_STATE_COMPLETED:
            job.completed = now
            job.state = consts.JOB_STATE_COMPLETED
            job.completion_status = consts.JOB_CMPLT_ALL_OK
            log.info('checking steps')
            for step in job.steps:
                log.info('%s: %s', step.index, consts.STEP_STATUS_NAME[step.status] if step.status in consts.STEP_STATUS_NAME else step.status)
                if step.status == consts.STEP_STATUS_ERROR:
                    # set base cmplt status error
                    job.completion_status = consts.JOB_CMPLT_AGENT_ERROR_RETURNED
                    # set proper cmplt status based on reason
                    if step.result and 'reason' in step.result:
                        if step.result['reason'] == 'job-timeout':
                            job.completion_status = consts.JOB_CMPLT_JOB_TIMEOUT
                        if step.result['reason'] == 'step-timeout':
                            job.completion_status = consts.JOB_CMPLT_STEP_TIMEOUT
                        elif step.result['reason'] == 'exception':
                            job.completion_status = consts.JOB_CMPLT_AGENT_EXCEPTION
                    break
            db.session.commit()

            _estimate_timeout(job)

        # establish new run state
        run = job.run
        is_completed = True
        for j in run.jobs:
            if j.state != consts.JOB_STATE_COMPLETED:
                is_completed = False
                break

        if is_completed:
            exec_utils.complete_run(run, now)


def _check_repo_commits(stage, flow_kind):
    triggers = stage.schema['triggers']
    if 'repo' not in triggers:
        return True, None  # no repo in triggers so start run; no repo_data

    # get prev run to get prev commit
    prev_run = dbutils.get_prev_run(stage.id, flow_kind)

    log.info('looking for new commits for stage %s, prev run %s', stage, prev_run)

    repo = triggers['repo']
    repos = []
    if 'url' in repo:
        repos.append((repo['url'], repo['branch']))
    else:
        for r in repo['repos']:
            repos.append((r['url'], r['branch']))

    git_cfg = repo.get('git_cfg', {})

    # iterate over repos
    changes = False
    repo_data = []
    for repo_url, repo_branch in repos:
        commits, base_commit = gitops.get_repo_commits_since(stage.branch_id, prev_run, repo_url, repo_branch, git_cfg)

        if commits:
            changes = True
            after = commits[0]['id']
        else:
            after = base_commit
        repo_data.append(dict(repo=repo_url,
                              commits=commits,
                              before=base_commit,
                              after=after,
                              trigger='git-push'))

    if not changes:
        log.info('no commits since prev check')
    else:
        log.info('detected commits since prev check')
    return changes, repo_data


def trigger_run(stage_id, flow_kind=consts.FLOW_KIND_CI, reason=None):
    log.reset_ctx()
    app = _create_app('trigger_run_%d' % stage_id)

    with app.app_context():
        stage = Stage.query.filter_by(id=stage_id).one_or_none()
        if stage is None:
            log.error('got unknown stage: %s', stage_id)
            return

        log.set_ctx(branch=stage.branch_id, flow_kind=flow_kind)
        log.info('triggering run for stage %s, flow kind %d', stage, flow_kind)

        # if this stage does not have parent then start new flow
        new_flow = False
        if stage.schema['parent'] == 'root':
            flow = Flow(branch=stage.branch, kind=flow_kind)
            new_flow = True

            # TODO: other root, sibling stages should be triggered when new flow is started

        else:
            # if this stage has parent then find latest flow with this parent stage run

            # first find parent stage using its name
            q = Stage.query.filter_by(name=stage.schema['parent'])
            q = q.filter_by(deleted=None)
            q = q.filter_by(branch_id=stage.branch_id)
            parent_stage = q.one_or_none()
            if parent_stage is None:
                log.error('parent stage %s for stage %s is missing', stage.schema['parent'], stage)
                return

            # find latest run of parent stage
            q = Run.query
            q = q.filter_by(deleted=None)
            q = q.filter_by(stage_id=parent_stage.id)
            q = q.join('flow')
            q = q.filter(Flow.kind == flow_kind)
            q = q.order_by(desc(Flow.created))
            parent_run = q.first()
            if parent_run is None:
                log.info('no run for parent stage %s', parent_stage)
                return

            # find if there is no run for current stage
            q = Run.query
            q = q.filter_by(deleted=None)
            q = q.filter_by(stage_id=stage.id)
            q = q.filter_by(flow_id=parent_run.flow_id)
            run = q.first()
            if run is not None:
                # TODO: in such case this trigger should be postponed to new flow
                log.info('latest flow %s with parent run %s already has run %s for current stage %s',
                         parent_run.flow_id, parent_run, run, stage)
                return

            flow = parent_run.flow

        # if the trigger is repo then first check if there were any new commits to the repo
        changes, repo_data = _check_repo_commits(stage, flow_kind)
        if not changes:
            return
        if repo_data:
            repo_data = RepoChanges(data=repo_data)
        if repo_data and reason and reason['reason'] == 'repo_interval':
            reason = dict(reason='commit to repo')
        if repo_data and new_flow:
            flow.trigger_data = repo_data

        # commit new flow if created and repo changes if detected
        db.session.commit()

        log.set_ctx(flow=flow.id)

        # create run for current stage
        if reason is None:
            reason = dict(reason='unknown')
        exec_utils.start_run(stage, flow, reason=reason, repo_data=repo_data)


def trigger_flow(project_id, trigger_data=None):
    log.reset_ctx()
    app = _create_app('trigger_flow_%d' % project_id)

    with app.app_context():
        project = Project.query.filter_by(id=project_id).one_or_none()
        if project is None:
            log.error('got unknown project: %s', project_id)
            return
        log.info('triggering flow for project %s', project)

        if trigger_data['trigger'] in ['github-push', 'gitea-push', 'gitlab-Push Hook']:
            branch_name = trigger_data['ref'].split('/')[-1]
            flow_kind = 'ci'
            flow_kind_no = 0
        elif trigger_data['trigger'] in ['github-pull_request', 'gitea-pull_request', 'gitlab-Merge Request Hook']:
            branch_name = trigger_data['pull_request']['base']['ref']
            flow_kind = 'dev'
            flow_kind_no = 1
        else:
            log.error('unsupported trigger %s', trigger_data['trigger'])
            return

        reason = trigger_data['trigger'].replace('-', ' ').replace('_', ' ')
        reason = reason.replace('Hook', '').strip()
        reason = dict(reason=reason)

        branch = Branch.query.filter_by(project=project, branch_name=branch_name).one_or_none()
        if branch is None:
            log.warning('cannot find branch by branch_name: %s', branch_name)
            return

        log.set_ctx(branch=branch.id)

        q = Flow.query
        q = q.filter_by(branch=branch)
        q = q.filter_by(kind=flow_kind_no)
        q = q.order_by(desc(Flow.created))
        last_flow = q.first()

        # handle the case when this is the first flow in the branch
        if last_flow is None:
            exec_utils.create_a_flow(branch, flow_kind, {})
            return

        trigger_git_url = trigger_data['repo']
        try:
            url = giturlparse.parse(trigger_git_url)
            trigger_git_url = url.url2https
        except Exception:
            log.warning('cannot parse trigger git url: %s', trigger_git_url)

        # find stages that use repo from trigger
        matching_stages = []
        for stage in branch.stages:
            if stage.deleted:
                continue
            found = False
            for job in stage.schema['jobs']:
                for step in job['steps']:
                    if step['tool'] == 'git':
                        git_url = step['checkout']
                        try:
                            url = giturlparse.parse(git_url)
                            git_url = url.url2https
                        except Exception:
                            # TODO: git url should be checked upfront and errors reported to user
                            log.warning('cannot parse git url: %s', git_url)
                            continue
                        if git_url == trigger_git_url:
                            matching_stages.append(stage)
                            found = True
                            break
                if found:
                    break

        # map runs to stages
        run_stages = {run.stage_id: run for run in last_flow.runs if not run.deleted}
        name_stages = {stage.name: stage for stage in branch.stages if not stage.deleted}

        # change trigger_data into db record
        trigger_data = RepoChanges(data=[trigger_data])

        # for stages that were not run while their parent was run, do run them now
        started_something = False
        for stage in matching_stages:
            if stage.id in run_stages:
                # TODO: trigger new flow?
                continue
            if not stage.enabled:
                log.info('stage %s not started - disabled', stage)
                continue
            parent_name = stage.schema['parent']
            if parent_name != 'root':
                parent_stage = name_stages[parent_name]
                if parent_stage.id not in run_stages:
                    # TODO: we should wait when parent is run and then trigger this stage
                    continue

            if not last_flow.trigger_data:
                last_flow.trigger_data = trigger_data
                db.session.commit()

            exec_utils.start_run(stage, last_flow, reason=reason)
            started_something = True

        if not started_something:
            exec_utils.create_a_flow(branch, flow_kind, {}, trigger_data)


def refresh_schema_repo(stage_id, complete_starting_run_id=None):
    log.reset_ctx()
    app = _create_app('refresh_schema_repo_%d' % stage_id)

    with app.app_context():
        stage = Stage.query.filter_by(id=stage_id).one_or_none()
        if stage is None:
            log.error('got unknown stage: %s', stage_id)
            return

        log.set_ctx(branch=stage.branch_id)
        run = None
        if complete_starting_run_id:
            run = Run.query.filter_by(id=complete_starting_run_id).one_or_none()
            if run is not None:
                log.set_ctx(flow_kind=run.flow.kind, flow=run.flow_id, run=run.id)

        log.info('refresh schema repo for stage: %d, run: %s',
                 stage, run)

        planner_url = os.environ.get('KRAKEN_PLANNER_URL', consts.DEFAULT_PLANNER_URL)
        planner = xmlrpc.client.ServerProxy(planner_url, allow_none=True)

        # cancel any previous scheduled refresh job
        if stage.repo_refresh_job_id:
            planner.remove_job(stage.repo_refresh_job_id)
            stage.repo_refresh_job_id = ''
            db.session.commit()

        try:
            # get schema from repo
            if stage.repo_access_token:
                secret = Secret.query.filter_by(project=stage.branch.project, name=stage.repo_access_token).one()
                repo_access_token = secret.data['secret']
            else:
                repo_access_token = None

            schema_code, version = gitops.get_schema_from_repo(stage.repo_url, stage.repo_branch, repo_access_token,
                                                               stage.schema_file, stage.git_clone_params)

            # check schema
            schema_code, schema = check_and_correct_stage_schema(stage.branch, stage.name, schema_code)
        except Exception as e:
            stage.repo_error = str(e)
            stage.repo_state = consts.REPO_STATE_ERROR
            db.session.commit()
            log.exception('problem with schema, stage: %d, run: %s',
                          stage_id, complete_starting_run_id)
            return

        # store schema id db
        prev_triggers = stage.schema['triggers']
        stage.schema = schema
        flag_modified(stage, 'schema')
        stage.schema_code = schema_code
        stage.repo_state = consts.REPO_STATE_OK
        stage.repo_error = ''
        stage.repo_version = version
        stage.schema_from_repo_enabled = True
        if stage.triggers is None:
            stage.triggers = {}
        prepare_new_planner_triggers(stage.id, schema['triggers'], prev_triggers, stage.triggers)
        flag_modified(stage, 'triggers')
        log.info('stage: %s, new schema: %s', stage, stage.schema)

        # start schema refresh job
        try:
            interval = int(stage.repo_refresh_interval)
        except Exception:
            interval = int(pytimeparse.parse(stage.repo_refresh_interval))
        job = planner.add_job('kraken.server.pljobs:refresh_schema_repo', 'interval', (stage_id,), None,
                              None, None, None, None, None, None, False, dict(seconds=interval))

        stage.repo_refresh_job_id = job['id']
        db.session.commit()

        if complete_starting_run_id:
            exec_utils.complete_starting_run(complete_starting_run_id)


def spawn_new_agents(agents_group_id):
    log.reset_ctx()
    app = _create_app('spawn_new_agents_%d' % agents_group_id)

    with app.app_context():
        server_url = get_setting('general', 'server_url')
        clickhouse_addr = get_setting('general', 'clickhouse_addr')
        settings = ['server_url', 'clickhouse_addr']
        for s in settings:
            val = locals()[s]
            if not val:
                log.error('%s is empty, please set it in global general settings', s)
                return

        ag = AgentsGroup.query.filter_by(id=agents_group_id).one_or_none()
        if ag is None:
            log.warning('cannot find agents group id: %d', agents_group_id)
            return

        try:
            _, depl = ag.get_deployment()
        except Exception:
            log.exception('IGNORED EXCEPTION')
            return

        # check if limit of agents is reached
        q = Agent.query.filter_by(authorized=True, disabled=False, deleted=None)
        q = q.join('agents_groups')
        q = q.filter_by(agents_group=ag)
        agents_count = q.count()
        limit = depl.get('instances_limit', 0)
        if agents_count >= limit:
            log.warning('in agents group id:%d cannot spawn more agents, limit %d reached', agents_group_id, limit)
            return

        # find waiting jobs assigned to this agents group
        jobs = defaultdict(int)
        q = Job.query.filter_by(covered=False, deleted=None, state=consts.JOB_STATE_QUEUED)
        q = q.filter_by(agents_group=ag)
        for job in q.all():
            sys_id = job.system_id if job.system.executor == 'local' else 0
            jobs[sys_id] += 1

        # go through systems needed and the counts and spawn agents
        for sys_id, needed_count in jobs.items():
            log.info('agents group %d, system id %d: needed %d', agents_group_id, sys_id, needed_count)

            # find system info
            if sys_id != 0:
                system = System.query.filter_by(id=sys_id).one_or_none()
                if system is None:
                    log.warning('cannot find system id:%d', sys_id)
                    continue
            else:
                system = System.query.filter_by(name=depl['default_image']).one_or_none()
                if system is None:
                    log.warning('cannot find system id:%d', sys_id)
                    continue

            # check if there is enough agents with proper system
            q = Agent.query.filter_by(authorized=True, disabled=False, deleted=None)
            q = q.join('agents_groups')
            q = q.filter_by(agents_group=ag)
            agents = q.all()
            agents_count = 0
            # log.info('potential agents num: %d', len(agents))
            for a in agents:
                if a.extra_attrs and 'system' in a.extra_attrs and a.extra_attrs['system'] == sys_id:
                    agents_count += 1
            num = needed_count - agents_count
            if num <= 0:
                log.info('enough agents, avail: %d, needed: %d', agents_count, needed_count)
                continue
            log.info('NOT enough agents, avail: %d, needed: %d, new: %d', agents_count, needed_count, num)

            cloud.create_machines(ag, system, num,
                                  server_url, clickhouse_addr)


def destroy_machine(agent_id):
    log.reset_ctx()
    app = _create_app('destroy_machine_%s' % agent_id)

    with app.app_context():
        agent = Agent.query.filter_by(id=int(agent_id)).one_or_none()
        if agent is None:
            log.error('cannot find agent id:%d', agent_id)
            return

        if not agent.extra_attrs:
            log.warning('missing extra_attrs in agent %s', agent)
            dbutils.delete_agent(agent)
            return

        ag = dbutils.find_cloud_assignment_group(agent)
        if not ag:
            log.error('agent %s does not have cloud group', agent)
            dbutils.delete_agent(agent)
            return

        dbutils.delete_agent(agent)

        cloud.destroy_machine(ag, agent)


def load_remote_tool(tool_id):
    log.reset_ctx()
    app = _create_app('load_remote_tool_%s' % tool_id)

    with app.app_context():
        tool_id = int(tool_id)
        tool = Tool.query.filter_by(id=tool_id).one_or_none()
        if tool is None:
            log.error('cannot find tool id:%d', tool_id)
            return

        # clone tool from remote repo
        try:
            tmpdir, repo_dir, version = gitops.clone_tool_repo(tool.url, tool.tag, tool_id)
        except Exception:
            log.exception('problem with cloning repo')
            return

        tf = None
        try:
            # package tool
            tool_file_path = os.path.join(repo_dir, tool.tool_file)
            if not os.path.exists(tool_file_path):
                log.warning('Tool meta file %s does not exist in repo %s', tool.tool_file, tool.url)
                return

            meta, tf, _ = toolops.package_tool(tool_file_path)

            name = meta['name']
            description = meta['description']
            location = meta['location']
            entry = meta['entry']
            fields = meta['parameters']

            # check if tool schema is ok
            toolutils.check_tool_schema(fields)

            if tool.version == tool.tag:
                # initial tool so set the fields from meta
                tool.name = name
                tool.description = description
                tool.version = version
                tool.location = location
                tool.entry = entry
                tool.fields = fields
            elif tool.version != version:
                # new version of tool so create new tool record in db
                tool = Tool(name=name, description=description, version=version, location=location, entry=entry, fields=fields)
                db.session.flush()
            elif tool.version == version:
                log.info('Tool %s name:%s from remote repo %s @ %s does not need to updated, the same version %s',
                         tool, tool.name, tool.url, tool.tag, version)

            # store tool
            toolutils.store_tool_in_minio(tf, tool)
            db.session.commit()

            log.info('Stored tool %s name:%s from remote repo %s @ %s', tool, tool.name, tool.url, tool.tag)

        finally:
            if tf:
                tf.close()
            tmpdir.cleanup()



# Copyright 2020-2021 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from queue import Queue, Empty
from threading import Thread, Event

from flask import abort, Response

from . import access
from . import users
from .models import Job
from . import chops


log = logging.getLogger(__name__)

class JobLogDownloader:
    def __init__(self, job_id, step_idx=None):
        self.job_id = job_id
        self.step_idx = step_idx
        self.query = 'select message from logs where job = %d ' % job_id
        if self.step_idx is not None:
            self.query += ' and step = %d ' % self.step_idx
        self.query += 'order by time asc, seq asc'

        self.ch = chops.get_clickhouse()

        self.logs_queue = Queue()
        self.finished = Event()
        self.worker = None

    def get_logs(self):
        for l in self.ch.execute_iter(self.query, {'max_block_size': 100000}):
            self.logs_queue.put(l[0] + '\n')

        self.logs_queue.join()  # wait for all blocks in the queue to be marked as processed
        self.finished.set()  # mark streaming as finished

    def send_logs(self):
        while not self.finished.is_set():
            try:
                yield self.logs_queue.get(timeout=0.01)
                self.logs_queue.task_done()
            except Empty:
                self.finished.wait(0.01)
        self.worker.join()

    def download(self, ):
        self.worker = Thread(target=self.get_logs)
        self.worker.start()
        return self.send_logs()


def serve_job_log(job_id):
    job_id = int(job_id)
    job = Job.query.filter_by(id=job_id).one_or_none()
    if job is None:
        abort(404, "Job not found")

    token_info = users.get_token_info_from_request()
    if not token_info:
        abort(401, "Missing Authorization header or kk_session_token cookie")
    access.check(token_info, job.run.flow.branch.project_id, 'view',
                 'only superadmin, project admin, project power user and project viewer roles can fetch job logs')

    try:
        jld = JobLogDownloader(job_id, None)
    except OSError:
        abort(500, 'Cannot connect to storage service')

    log.info('serve_job_log %s', job)
    response = Response(jld.download(), mimetype='plain/text')
    path = 'job_log_%d.txt' % job.id
    response.headers['Content-Disposition'] = 'attachment; filename=' + path
    return response


def serve_step_log(job_id, step_idx):
    job_id = int(job_id)
    step_idx = int(step_idx)
    job = Job.query.filter_by(id=job_id).one_or_none()
    if job is None:
        abort(404, "Job not found")

    token_info = users.get_token_info_from_request()
    if not token_info:
        abort(401, "Missing Authorization header or kk_session_token cookie")
    access.check(token_info, job.run.flow.branch.project_id, 'view',
                 'only superadmin, project admin, project power user and project viewer roles can fetch job logs')

    try:
        jld = JobLogDownloader(job_id, step_idx)
    except OSError:
        abort(500, 'Cannot connect to storage service')

    log.info('serve_job_log %s', job)
    response = Response(jld.download(), mimetype='plain/text')
    path = 'step_log_%d_%d.txt' % (job_id, step_idx)
    response.headers['Content-Disposition'] = 'attachment; filename=' + path
    return response



"""uniqued agent address

Revision ID: b84bdc5d3c67
Revises: b2c9024746f5
Create Date: 2021-06-21 22:03:17.503080

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = 'b84bdc5d3c67'
down_revision = 'b2c9024746f5'
branch_labels = None
depends_on = None


def upgrade():
    op.drop_index('ix_agents_address', table_name='agents')
    op.create_index(op.f('ix_agents_address'), 'agents', ['address'], unique=True)


def downgrade():
    op.drop_index(op.f('ix_agents_address'), table_name='agents')
    op.create_index('ix_agents_address', 'agents', ['address'], unique=False)



"""added artifacts to Flow

Revision ID: 2b5751b83fec
Revises: 66808ea51a72
Create Date: 2020-02-12 08:30:28.025708

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '2b5751b83fec'
down_revision = '66808ea51a72'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('flows', sa.Column('artifacts', postgresql.JSONB(astext_type=sa.Text()), nullable=True))


def downgrade():
    op.drop_column('flows', 'artifacts')



# Copyright 2020 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

LOG_FMT = '%(asctime)s %(levelname)-4.4s p:%(process)5d %(module)8.8s:%(lineno)-5d %(message)s'
TOOL_LOG_FMT = '%(message)s'

JOB_STATE_PREQUEUED = 1
JOB_STATE_QUEUED = 2
JOB_STATE_ASSIGNED = 3
JOB_STATE_EXECUTING_FINISHED = 4
JOB_STATE_COMPLETED = 5

JOB_STATES_NAME = {
    JOB_STATE_PREQUEUED: "prequeued",
    JOB_STATE_QUEUED: "queued",
    JOB_STATE_ASSIGNED: "assigned",
    JOB_STATE_EXECUTING_FINISHED: "executing-finished",
    JOB_STATE_COMPLETED: "completed",
}

JOB_CMPLT_ALL_OK = 0
JOB_CMPLT_JOB_TIMEOUT = 1
JOB_CMPLT_AGENT_ERROR_RETURNED = 2
JOB_CMPLT_AGENT_EXCEPTION = 3
JOB_CMPLT_MISSING_TOOL_IN_DB = 4
JOB_CMPLT_MISSING_TOOL_FILES = 5
JOB_CMPLT_STEP_TIMEOUT = 6
JOB_CMPLT_SERVER_TIMEOUT = 7
JOB_CMPLT_USER_CANCEL = 8
JOB_CMPLT_MISSING_AGENTS_GROUP = 9
JOB_CMPLT_NO_AGENTS = 10
JOB_CMPLT_AGENT_NOT_ALIVE = 11

STEP_STATUS_NOT_STARTED = 0
STEP_STATUS_IN_PROGRES = 1
STEP_STATUS_DONE = 2
STEP_STATUS_ERROR = 3

STEP_STATUS_TO_INT = {
    'not-started': STEP_STATUS_NOT_STARTED,
    'in-progress': STEP_STATUS_IN_PROGRES,
    'done': STEP_STATUS_DONE,
    'error': STEP_STATUS_ERROR
}
STEP_STATUS_NAME = {
    STEP_STATUS_NOT_STARTED: 'not-started',
    STEP_STATUS_IN_PROGRES: 'in-progress',
    STEP_STATUS_DONE: 'done',
    STEP_STATUS_ERROR: 'error'
}

TC_RESULT_NOT_RUN = 0
TC_RESULT_PASSED = 1
TC_RESULT_FAILED = 2
TC_RESULT_ERROR = 3
TC_RESULT_DISABLED = 4
TC_RESULT_UNSUPPORTED = 5
TC_RESULTS_NAME = {
    TC_RESULT_NOT_RUN: 'Not Run',
    TC_RESULT_PASSED: 'Passed',
    TC_RESULT_FAILED: 'Failed',
    TC_RESULT_ERROR: 'Error',
    TC_RESULT_DISABLED: 'Disabled',
    TC_RESULT_UNSUPPORTED: 'Unsupported',
}

TC_COMMENT_NEW = 0
TC_COMMENT_INVESTIGATING = 1
TC_COMMENT_BUG_IN_PRODUCT = 2
TC_COMMENT_BUG_IN_TEST = 3

RUN_STATE_IN_PROGRESS = 1
RUN_STATE_COMPLETED = 2
RUN_STATE_PROCESSED = 3
RUN_STATE_MANUAL = 4
RUN_STATE_REPLAY = 5
RUN_STATE_CREATING = 6

RUN_STATES_NAME = {
    RUN_STATE_IN_PROGRESS: 'in-progress',
    RUN_STATE_COMPLETED: 'completed',
    RUN_STATE_PROCESSED: 'processed',
    RUN_STATE_MANUAL: 'manual',
    RUN_STATE_REPLAY: 'replay',
    RUN_STATE_CREATING: 'creating',
}

FLOW_KIND_CI = 0
FLOW_KIND_DEV = 1

FLOW_STATE_IN_PROGRESS = 1
FLOW_STATE_COMPLETED = 2

FLOW_STATES_NAME = {
    FLOW_STATE_IN_PROGRESS: 'in-progress',
    FLOW_STATE_COMPLETED: 'completed'
}

TC_RESULT_CHANGE_NO = 0
TC_RESULT_CHANGE_FIX = 1
TC_RESULT_CHANGE_REGR = 2
TC_RESULT_CHANGE_NEW = 3

TC_RESULT_CHANGES_NAME = {
    TC_RESULT_CHANGE_NO: 'no-change',
    TC_RESULT_CHANGE_FIX: 'fix',
    TC_RESULT_CHANGE_REGR: 'regr',
    TC_RESULT_CHANGE_NEW: 'new'
}


NETWORK_TIMEOUT = 2  # minutes


DEFAULT_DB_URL = 'postgresql://kraken:kk123@localhost:5433/kraken'
DEFAULT_REDIS_ADDR = 'localhost:6379'
DEFAULT_CLICKHOUSE_PORT = '9001'
DEFAULT_CLICKHOUSE_ADDR = 'localhost:%s' % DEFAULT_CLICKHOUSE_PORT
DEFAULT_CLICKHOUSE_URL = 'http://localhost:8123'
DEFAULT_PLANNER_URL = 'http://localhost:7997/'
DEFAULT_SERVER_ADDR = 'localhost:8080'
DEFAULT_MINIO_ADDR = 'localhost:9999'

ISSUE_TYPE_ERROR = 0
ISSUE_TYPE_WARNING = 1
ISSUE_TYPE_CONVENTION = 2
ISSUE_TYPE_REFACTOR = 3
ISSUE_TYPES_NAME = {
    ISSUE_TYPE_ERROR: 'error',
    ISSUE_TYPE_WARNING: 'warning',
    ISSUE_TYPE_CONVENTION: 'convention',
    ISSUE_TYPE_REFACTOR: 'refactor',
}
ISSUE_TYPES_CODE = {n: c for c, n in ISSUE_TYPES_NAME.items()}

SECRET_KIND_SSH_KEY = 0
SECRET_KIND_SIMPLE = 1
SECRET_KINDS_NAME = {
    SECRET_KIND_SSH_KEY: 'ssh-key',
    SECRET_KIND_SIMPLE: 'simple',
}


DEFAULT_RUN_TIMEOUT = 3 * 60 * 60  # 3 hours (in seconds)
DEFAULT_JOB_TIMEOUT = 5 * 60  # 5 minutes (in seconds)
AGENT_TIMEOUT = 5 * 60  # 5 minutes (in seconds)
SLOW_AGENT_TIMEOUT = 10 * 60  # 10 minutes (in seconds)


ARTIFACTS_SECTION_PRIVATE = 0
ARTIFACTS_SECTION_PUBLIC = 1


AGENT_DIR = '/opt/kraken'


AGENT_MSG_GET_JOB = 'get-job'
AGENT_MSG_STEP_RESULT = 'step-result'
AGENT_MSG_DISPATCH_TESTS = 'dispatch-tests'
AGENT_MSG_HOST_INFO = 'host-info'
AGENT_MSG_KEEP_ALIVE = 'keep-alive'


BRANCH_SEQ_FLOW = 0
BRANCH_SEQ_CI_FLOW = 1
BRANCH_SEQ_DEV_FLOW = 2
BRANCH_SEQ_RUN = 3
BRANCH_SEQ_CI_RUN = 4
BRANCH_SEQ_DEV_RUN = 5


REPO_STATE_OK = 0
REPO_STATE_REFRESHING = 1
REPO_STATE_ERROR = 2

AGENT_DEPLOYMENT_METHOD_MANUAL = 0
#AGENT_DEPLOYMENT_METHOD_SSH = 1
AGENT_DEPLOYMENT_METHOD_AWS_EC2 = 2
AGENT_DEPLOYMENT_METHOD_AWS_ECS_FARGATE = 3
AGENT_DEPLOYMENT_METHOD_AZURE_VM = 4
AGENT_DEPLOYMENT_METHOD_K8S = 5
#AGENT_DEPLOYMENT_METHOD_GCP =
#AGENT_DEPLOYMENT_METHOD_DO =


REDIS_KRAKEN_DB = 1
REDIS_RQ_DB = 2



# Copyright 2020 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import smtplib
from urllib.parse import urljoin
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart

import requests

from .models import get_setting
from .schema import prepare_secrets, substitute_vars


log = logging.getLogger(__name__)


SLACK_URL = 'https://slack.com/api/chat.postMessage'


def _get_srv_url():
    server_url = get_setting('general', 'server_url')
    if server_url is None:
        server_url = 'http://localhost:4200'
    return server_url


def _notify_slack(run, event, slack):
    if event == 'start':
        return
    if slack is None:
        return

    channel = slack.get('channel', None)
    if channel is None:
        return

    slack_token = get_setting('notification', 'slack_token')
    if slack_token is None:
        return

    server_url = _get_srv_url()

    text = 'Project <%s|%s>, branch <%s|%s> <%s|%s>, flow <%s|%s>, run <%s|%s %s>'
    text = text % (urljoin(server_url, '/projects/%d' % run.flow.branch.project.id),
                   run.flow.branch.project.name,
                   urljoin(server_url, '/branches/%s' % run.flow.branch.id),
                   run.flow.branch.name,
                   urljoin(server_url, '/branches/%s/%s' % (run.flow.branch.id, 'ci' if run.flow.kind == 0 else 'dev')),
                   'CI' if run.flow.kind == 0 else 'Dev',
                   urljoin(server_url, '/flows/%d' % run.flow.id),
                   run.flow.id,
                   urljoin(server_url, '/runs/%d' % run.id),
                   run.stage.name,
                   run.id)
    if run.regr_cnt > 0:
        text += ', regressions: %d' % run.regr_cnt
    if run.fix_cnt > 0:
        text += ', fixes: %d' % run.fix_cnt
    if run.issues_new > 0:
        text += ', new issues: %d' % run.issues_new

    log.info('slack channel: %s, msg: %s', channel, text)
    data = dict(token=slack_token, channel=channel, text=text)

    # sending post to slack
    r = requests.post(SLACK_URL, data=data)

    log.info('slack resp: %s, %s', r, r.text)


def _notify_email(run, event, email):
    if event == 'start':
        return
    if email is None:
        return
    recipients = email

    smtp_server = get_setting('notification', 'smtp_server')
    if smtp_server is None:
        return

    if ':' in smtp_server:
        smtp_server, smtp_port = smtp_server.split(':')
        smtp_port = int(smtp_port)
    else:
        smtp_port = 0
    smtp_user = get_setting('notification', 'smtp_user')
    smtp_password = get_setting('notification', 'smtp_password')
    smtp_from = get_setting('notification', 'smtp_from')
    if smtp_from is None:
        smtp_from = 'kraken@kraken'
    smtp_tls = get_setting('notification', 'smtp_tls')

    server_url = _get_srv_url()

    html = 'Project <a href="%s">%s</a>, branch <a href="%s">%s</a> <a href="%s">%s</a>, flow <a href="%s">%s</a>, run <a href="%s">%s %s</a>'
    html = html % (urljoin(server_url, '/projects/%d' % run.flow.branch.project.id),
                   run.flow.branch.project.name,
                   urljoin(server_url, '/branches/%s' % run.flow.branch.id),
                   run.flow.branch.name,
                   urljoin(server_url, '/branches/%s/%s' % (run.flow.branch.id, 'ci' if run.flow.kind == 0 else 'dev')),
                   'CI' if run.flow.kind == 0 else 'Dev',
                   urljoin(server_url, '/flows/%d' % run.flow.id),
                   run.flow.id,
                   urljoin(server_url, '/runs/%d' % run.id),
                   run.stage.name,
                   run.id)
    html += '<br>'
    if run.tests_total > 0:
        html += '<br><b>Tests</b><br>'
        html += 'regressions: %d<br>' % run.regr_cnt
        html += 'fixes: %d<br>' % run.fix_cnt
        html += 'passed: %d<br>' % run.tests_passed
        html += 'total: %d<br>' % run.tests_total
        html += 'new: %d<br>' % run.new_cnt

    if run.issues_total > 0:
        html += '<br><b>Issues</b><br>'
        html += 'new issues: %d<br>' % run.issues_new
        html += 'total: %d<br>' % run.issues_total

    html += '<br>-- Kraken'

    subject = '[Kraken] Project %s, branch %s %s, flow %s, run %s %s'
    subject = subject % (run.flow.branch.project.name,
                         run.flow.branch.name,
                         'CI' if run.flow.kind == 0 else 'Dev',
                         run.flow.id,
                         run.stage.name,
                         run.id)
    if run.regr_cnt > 0:
        subject += ', regressions: %d' % run.regr_cnt
    if run.fix_cnt > 0:
        subject += ', fixes: %d' % run.fix_cnt
    if run.issues_new > 0:
        subject += ', new issues: %d' % run.issues_new

    message = MIMEMultipart("alternative")
    message["Subject"] = subject
    message["From"] = smtp_from
    message["To"] = recipients
    part = MIMEText(html, "html")
    message.attach(part)

    # sending email to smtp server
    if smtp_tls:
        server = smtplib.SMTP_SSL(smtp_server, smtp_port)
    else:
        server = smtplib.SMTP(smtp_server, smtp_port)
        server.ehlo()
    if smtp_user and smtp_password:
        server.login(smtp_user, smtp_password)


    log.info("sending email, server: '%s:%s', TLS: %s, user: '%s', password: '%s'",
             smtp_server, str(smtp_port),
             'yes' if smtp_tls else 'no',
             smtp_user if smtp_user else '',
             '*' * len(smtp_password) if smtp_password else '')

    server.sendmail(smtp_from, recipients, message.as_string())

    server.close()

    log.info('email sent')


def _notify_github(run, event, gh):
    if gh is None:
        return
    if not run.flow.trigger_data:
        return

    # prepare github credentials
    creds = gh.get('credentials', None)
    if creds is None:
        log.error('no github credentials')
        return
    creds = creds.split(':')
    if len(creds) != 2:
        log.error('github credentials should have user:token form')
        return
    creds = tuple(creds)  # requests require tuple, not list

    # prepare data for github status
    context = 'kraken / %s [%s]' % (run.stage.name, run.flow.get_label())

    if event == 'start':
        state = 'pending'
        descr = 'waiting for results'
    elif event == 'end':
        state = 'success'
        descr = []
        if run.regr_cnt > 0:
            descr.append('regressions: %d' % run.regr_cnt)
            state = 'failure'
        if run.fix_cnt > 0:
            descr.append('fixes: %d' % run.fix_cnt)
        if run.issues_new > 0:
            descr.append('new issues: %d' % run.issues_new)
            state = 'failure'
        descr = ', '.join(descr)
    else:
        log.error('unsupported event %s', event)
        return

    server_url = _get_srv_url()
    run_url = urljoin(server_url, '/runs/%d' % run.id)

    head = run.flow.trigger_data.data[0]['after']
    repo_parts = run.flow.trigger_data.data[0]['repo'].split('/')
    org = repo_parts[-2]
    repo_name = repo_parts[-1]
    if repo_name.endswith('.git'):
        repo_name = repo_name[:-4]
    url = 'https://api.github.com/repos/%s/%s/statuses/%s' % (org, repo_name, head)

    data = {
        'state': state,
        'context': context,
        'target_url': run_url,
        'description': descr
    }

    log.info('GH data %s', data)
    r = requests.post(url, data=json.dumps(data), auth=creds)

    log.info('github resp: %s, %s', r, r.text)


def notify(run, event):
    notification = run.stage.schema.get('notification', None)
    if notification is None:
        return

    # prepare secrets to pass them to substitute in notifications definitions
    args = prepare_secrets(run)
    args.update(run.args)
    # log.info('notification1 %s', notification)
    notification, _ = substitute_vars(notification, args)
    # log.info('notification2 %s', notification)

    # slack
    slack = notification.get('slack', None)
    try:
        _notify_slack(run, event, slack)
    except Exception:
        log.exception('IGNORED EXCEPTION')

    # email
    email = notification.get('email', None)
    try:
        _notify_email(run, event, email)
    except Exception:
        log.exception('IGNORED EXCEPTION')

    # github
    github = notification.get('github', None)
    try:
        _notify_github(run, event, github)
    except Exception:
        log.exception('IGNORED EXCEPTION')


def check_email_settings():
    smtp_server = get_setting('notification', 'smtp_server')
    if not smtp_server:
        return 'STMP server address is empty'

    if ':' in smtp_server:
        smtp_server, smtp_port = smtp_server.split(':')
        smtp_port = int(smtp_port)
    else:
        smtp_port = 0
    smtp_user = get_setting('notification', 'smtp_user')
    smtp_password = get_setting('notification', 'smtp_password')
    smtp_tls = get_setting('notification', 'smtp_tls')
    try:
        if smtp_tls:
            server = smtplib.SMTP_SSL(smtp_server, smtp_port)
        else:
            server = smtplib.SMTP(smtp_server, smtp_port)
            server.ehlo()
        if smtp_user and smtp_password:
            server.login(smtp_user, smtp_password)
    except Exception as ex:
        return str(ex)

    return 'ok'


def check_slack_settings():
    slack_token = get_setting('notification', 'slack_token')
    if not slack_token:
        return 'Slack token is empty'

    # TODO

    return 'ok'



"""added reason to run

Revision ID: 29b96da53f09
Revises: dae910bd9587
Create Date: 2021-01-10 13:41:43.387949

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '29b96da53f09'
down_revision = 'dae910bd9587'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('runs', sa.Column('reason', postgresql.JSONB(astext_type=sa.Text()), nullable=True))

    conn = op.get_bind()
    conn.execute("""UPDATE runs SET reason = '{"reason": "manual"}'::jsonb""")

    op.alter_column('runs', 'reason', nullable=False)


def downgrade():
    op.drop_column('runs', 'reason')



"""added branch sequences

Revision ID: 4a7ac9143bb1
Revises: 38088514157a
Create Date: 2020-09-26 11:18:26.316516

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '4a7ac9143bb1'
down_revision = '38088514157a'
branch_labels = None
depends_on = None


def upgrade():
    bs_tbl = op.create_table('branch_sequences',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('kind', sa.Integer(), nullable=True),
                    sa.Column('branch_id', sa.Integer(), nullable=False),
                    sa.Column('stage_id', sa.Integer(), nullable=True),
                    sa.Column('value', sa.Integer(), nullable=True),
                    sa.ForeignKeyConstraint(['branch_id'], ['branches.id'], ),
                    sa.ForeignKeyConstraint(['stage_id'], ['stages.id'], ),
                    sa.PrimaryKeyConstraint('id')
    )
    conn = op.get_bind()
    br_res = conn.execute("SELECT id FROM branches WHERE deleted IS NULL")
    seqs = []
    for br_id in br_res.fetchall():
        seqs.append({'kind': 0, 'branch_id': br_id[0], 'stage_id': None, 'value': 0})
        seqs.append({'kind': 1, 'branch_id': br_id[0], 'stage_id': None, 'value': 0})
        seqs.append({'kind': 2, 'branch_id': br_id[0], 'stage_id': None, 'value': 0})

        stg_res = conn.execute("SELECT id FROM stages WHERE deleted IS NULL AND branch_id = %d" % br_id[0])
        for stg_id in stg_res.fetchall():
            seqs.append({'kind': 3, 'branch_id': br_id[0], 'stage_id': stg_id[0], 'value': 0})
            seqs.append({'kind': 4, 'branch_id': br_id[0], 'stage_id': stg_id[0], 'value': 0})
            seqs.append({'kind': 5, 'branch_id': br_id[0], 'stage_id': stg_id[0], 'value': 0})

    op.bulk_insert(bs_tbl, seqs)


def downgrade():
    op.drop_table('branch_sequences')



"""update Stage.enabled to true

Revision ID: ba436d491bf8
Revises: 0731897c862e
Create Date: 2019-12-22 19:00:15.365002

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = 'ba436d491bf8'
down_revision = '0731897c862e'
branch_labels = None
depends_on = None


def upgrade():
    conn = op.get_bind()
    conn.execute("UPDATE stages SET enabled = TRUE")


def downgrade():
    pass



"""rename executor to agent

Revision ID: 89ffd52f1f00
Revises: 6745adaf40e5
Create Date: 2020-09-05 08:19:47.642453

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = '89ffd52f1f00'
down_revision = '6745adaf40e5'
branch_labels = None
depends_on = None


def upgrade():

    op.rename_table("executors", "agents")
    op.rename_table("executor_assignments", "agent_assignments")
    op.rename_table("executor_groups", "agents_groups")

    op.drop_constraint("executor_assignments_executor_id_fkey", "agent_assignments", type_="foreignkey")
    op.drop_constraint("executor_assignments_executor_group_id_fkey", "agent_assignments", type_="foreignkey")
    op.drop_constraint("executor_assignments_pkey", "agent_assignments", type_="primary")
    op.drop_constraint("jobs_executor_used_id_fkey", "jobs", type_="foreignkey")
    op.drop_constraint("jobs_executor_group_id_fkey", "jobs", type_="foreignkey")
    op.drop_constraint("executors_pkey", "agents", type_="primary")
    op.drop_constraint("executors_job_id_fkey", "agents", type_="foreignkey")
    op.drop_constraint("executor_groups_pkey", "agents_groups", type_="primary")
    op.drop_constraint("executor_groups_project_id_fkey", "agents_groups", type_="primary")

    op.alter_column("agent_assignments", "executor_id", new_column_name="agent_id")
    op.alter_column("agent_assignments", "executor_group_id", new_column_name="agents_group_id")
    op.alter_column("jobs", "executor_group_id", new_column_name="agents_group_id")
    op.alter_column("jobs", "executor_used_id", new_column_name="agent_used_id")

    op.create_primary_key("agents_pkey", "agents", ["id"])
    op.create_primary_key("agent_assignments_pkey", "agent_assignments", ["agent_id", "agents_group_id"])
    op.create_primary_key("agents_groups_pkey", "agents_groups", ["id"])
    op.create_foreign_key("jobs_agent_used_id_fkey", "jobs",
                          "agents", ["agent_used_id"], ["id"])
    op.create_foreign_key("jobs_agents_group_id_fkey", "jobs",
                          "agents_groups", ["agents_group_id"], ["id"])
    op.create_foreign_key("agent_assignments_agent_id_fkey", "agent_assignments",
                          "agents", ["agent_id"], ["id"])
    op.create_foreign_key("agent_assignments_agents_group_id_fkey", "agent_assignments",
                          "agents_groups", ["agents_group_id"], ["id"])
    op.create_foreign_key("agents_groups_project_id_fkey", "agents_groups",
                          "projects", ["project_id"], ["id"])
    op.create_foreign_key("agents_job_id_fkey", "agents",
                          "jobs", ["job_id"], ["id"])

    op.drop_index("ix_executors_address")
    op.create_index('ix_agents_address', 'agents', ['address'])

    op.execute('ALTER SEQUENCE executor_groups_id_seq RENAME TO agents_groups_id_seq')
    op.execute('ALTER SEQUENCE executors_id_seq RENAME TO agents_id_seq')


def downgrade():
    op.rename_table("agent_assignments", "executor_assignments")
    op.rename_table("agents", "executors")
    op.rename_table("agents_groups", "executor_groups")

    op.drop_constraint("agent_assignments_agent_id_fkey", "executor_assignments", type_="foreignkey")
    op.drop_constraint("agent_assignments_agents_group_id_fkey", "executor_assignments", type_="foreignkey")
    op.drop_constraint("agent_assignments_pkey", "executor_assignments", type_="primary")
    op.drop_constraint("jobs_agent_used_id_fkey", "jobs", type_="foreignkey")
    op.drop_constraint("jobs_agents_group_id_fkey", "jobs", type_="foreignkey")
    op.drop_constraint("agents_pkey", "executors", type_="primary")
    op.drop_constraint("agents_job_id_fkey", "executors", type_="foreignkey")
    op.drop_constraint("agents_groups_pkey", "executor_groups", type_="primary")
    op.drop_constraint("agents_groups_project_id_fkey", "executor_groups", type_="primary")

    op.alter_column("executor_assignments", "agents_group_id", new_column_name="executor_group_id")
    op.alter_column("executor_assignments", "agent_id", new_column_name="executor_id")
    op.alter_column("jobs", "agent_used_id", new_column_name="executor_used_id")
    op.alter_column("jobs", "agents_group_id", new_column_name="executor_group_id")

    op.create_primary_key("executors_pkey", "executors", ["id"])
    op.create_primary_key("executor_assignments_pkey", "executor_assignments", ["executor_id", "executor_group_id"])
    op.create_primary_key("executor_groups_pkey", "executor_groups", ["id"])
    op.create_foreign_key("jobs_executor_used_id_fkey", "jobs",
                          "executors", ["executor_used_id"], ["id"])
    op.create_foreign_key("jobs_executor_group_id_fkey", "jobs",
                          "executor_groups", ["executor_group_id"], ["id"])
    op.create_foreign_key("executor_assignments_executor_id_fkey", "executor_assignments",
                          "executors", ["executor_id"], ["id"])
    op.create_foreign_key("executor_assignments_executor_group_id_fkey", "executor_assignments",
                          "executor_groups", ["executor_group_id"], ["id"])
    op.create_foreign_key("executor_groups_project_id_fkey", "executor_groups",
                          "projects", ["project_id"], ["id"])
    op.create_foreign_key("executors_job_id_fkey", "executors",
                          "jobs", ["job_id"], ["id"])

    op.drop_index("ix_agents_address")
    op.create_index('ix_executors_address', 'executors', ['address'])

    op.execute('ALTER SEQUENCE agents_groups_id_seq RENAME TO executor_groups_id_seq')
    op.execute('ALTER SEQUENCE agents_id_seq RENAME TO executors_id_seq')



"""added label to flow

Revision ID: d2b23f410d02
Revises: 4a7ac9143bb1
Create Date: 2020-09-26 17:17:27.078419

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'd2b23f410d02'
down_revision = '4a7ac9143bb1'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('flows', sa.Column('label', sa.Unicode(length=200), nullable=True))


def downgrade():
    op.drop_column('flows', 'label')



"""delete Setting records with no group

Revision ID: 66808ea51a72
Revises: 8b4adeca953d
Create Date: 2020-02-09 06:35:05.418249

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = '66808ea51a72'
down_revision = '8b4adeca953d'
branch_labels = None
depends_on = None


def upgrade():
    conn = op.get_bind()
    conn.execute("DELETE FROM settings where settings.group = NULL;")

def downgrade():
    pass



"""added indexes to test_case_results

Revision ID: a0741baf2dbc
Revises: f566580fd139
Create Date: 2022-01-20 06:50:44.181743

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = 'a0741baf2dbc'
down_revision = 'f566580fd139'
branch_labels = None
depends_on = None


def upgrade():
    op.create_index('ix_test_case_results_comment_id', 'test_case_results', ['comment_id'], unique=False)
    op.create_index('ix_test_case_results_job_id', 'test_case_results', ['job_id'], unique=False)


def downgrade():
    op.drop_index('ix_test_case_results_job_id', table_name='test_case_results')
    op.drop_index('ix_test_case_results_comment_id', table_name='test_case_results')



"""added url, tag and tool_file to tool

Revision ID: 0615faca43b1
Revises: befa07ac5d1c
Create Date: 2022-05-30 20:34:29.878224

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '0615faca43b1'
down_revision = 'befa07ac5d1c'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('tools', sa.Column('url', sa.UnicodeText(), nullable=True))
    op.add_column('tools', sa.Column('tag', sa.UnicodeText(), nullable=True))
    op.add_column('tools', sa.Column('tool_file', sa.UnicodeText(), nullable=True))


def downgrade():
    op.drop_column('tools', 'tool_file')
    op.drop_column('tools', 'tag')
    op.drop_column('tools', 'url')



"""added extra_atrs to agents

Revision ID: a26a02d0ac01
Revises: 050fbc3d126c
Create Date: 2021-05-05 06:32:20.203580

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'a26a02d0ac01'
down_revision = '050fbc3d126c'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('agents', sa.Column('extra_attrs', postgresql.JSONB(astext_type=sa.Text()), nullable=True))


def downgrade():
    op.drop_column('agents', 'extra_attrs')



# Copyright 2021-2022 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import logging

import redis
from flask import abort, redirect, Response

from . import consts
from . import utils


log = logging.getLogger(__name__)


def _redir_to_badge_image(rds, branch_id, what):
    if what is None:
        label = 'Kraken Build'
    elif what == 'tests':
        label = 'Kraken Tests'
    elif what == 'issues':
        label = 'Kraken Issues'
    else:
        raise Exception('not supported badge category %s' % what)

    # get data from redis
    # data in redis is prepared in bg/jobs.py, in _prepare_flow_summary()
    key = 'branch-%s' % branch_id
    data = rds.get(key)
    if not data:
        url = 'https://img.shields.io/badge/%s-%s-%s' % (label, 'no data', 'inactive')
        return redirect(url)

    data = json.loads(data)

    msg = '%s ' % data['label']

    if what is None:
        # flow status
        if data['errors']:
            color = 'critical'
            msg += 'failed'
        else:
            color = 'success'
            msg += 'success'

    elif what == 'tests':
        # flow tests
        if data['tests_total'] == 0:
            msg += 'no tests'
            color = 'informational'
        else:
            color = 'success'
            if data['tests_passed'] < data['tests_total']:
                color = 'important'
            pass_ratio = 100 * data['tests_passed'] / data['tests_total']
            msg += 'passed %.1f%%25 total %d' % (pass_ratio, data['tests_total'])
            if data['tests_regr'] > 0:
                msg += ' regressions %d' % data['tests_regr']
            if data['tests_fix'] > 0:
                msg += ' fixes %d' % data['tests_fix']

    elif what == 'issues':
        # flow issues
        color = 'success'
        if data['issues_total'] == 0:
            msg += 'no issues'
        else:
            msg += 'issues %d' % data['issues_total']
            color = 'important'
        if data['issues_new'] > 0:
            msg += 'new %d' % data['issues_new']
            color = 'critical'

    url = 'https://img.shields.io/badge/%s-%s-%s' % (label, msg, color)
    return redirect(url)


def _get_cctray_status(rds, branch_id):
    # get data from redis
    # data in redis is prepared in bg/jobs.py, in _prepare_flow_summary()
    key = 'branch-%s' % branch_id
    data = rds.get(key)
    if data:
        data = json.loads(data)
    else:
        data = dict(project='unknown',
                    branch='unknown',
                    activity='Sleeping',
                    label='unknown',
                    lastBuildTime='2005-09-28T10:30:34.6362160+01:00',
                    url='http://example.com',
                    errors=False)

    if data['errors']:
        data['status'] = 'Failure'
    else:
        data['status'] = 'Success'

    xml = '''<?xml version="1.0" encoding="utf-8"?>
    <Projects>
      <Project
        name="{project} - {branch}"
        activity="{activity}"
        lastBuildStatus="{status}"
        lastBuildLabel="{label}"
        lastBuildTime="{lastBuildTime}"
        webUrl="{url}"/>
    </Projects>'''

    xml = xml.format(**data)

    r = Response(response=xml, status=200, mimetype="application/xml")
    r.headers["Content-Type"] = "text/xml; charset=utf-8"
    return r


def get_branch_badge(branch_id, what=None):
    # get redis reference
    redis_addr = os.environ.get('KRAKEN_REDIS_ADDR', consts.DEFAULT_REDIS_ADDR)
    redis_host, redis_port = utils.split_host_port(redis_addr, 6379)
    rds = redis.Redis(host=redis_host, port=redis_port, db=consts.REDIS_KRAKEN_DB)

    try:
        int(branch_id)
    except Exception:
        abort(400, 'wrong branch id')

    if what == 'cctray':
        return _get_cctray_status(rds, branch_id)
    return _redir_to_badge_image(rds, branch_id, what)



"""longer message in Issue

Revision ID: b908e219f07f
Revises: 1982be95e79f
Create Date: 2020-01-26 07:59:37.454201

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'b908e219f07f'
down_revision = '1982be95e79f'
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column('issues', 'message',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Unicode(length=512),
               existing_nullable=True)


def downgrade():
    op.alter_column('issues', 'message',
               existing_type=sa.Unicode(length=512),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
